# RCAIDE/Library/Methods/Aerodynamics/Vortex_Lattice_Method/build_VLM_surrogates.py
#  
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
from RCAIDE.Framework.Core import  Data 

# package imports 
from scipy.interpolate                                           import RegularGridInterpolator
from scipy import interpolate

# ----------------------------------------------------------------------------------------------------------------------
#  Vortex_Lattice
# ----------------------------------------------------------------------------------------------------------------------   
def build_VLM_surrogates(aerodynamics):
    """Build a surrogate using sample evaluation results.
    
    Assumptions:
        None
        
    Source:
        None

    Args:
        aerodynamics       : VLM analysis          [unitless] 
        
    Returns: 
        None  
    """
    surrogates =  aerodynamics.surrogates
    training   =  aerodynamics.training 
    surrogates.subsonic    =  build_surrogate(aerodynamics, training.subsonic) 
    surrogates.supersonic  =  build_surrogate(aerodynamics, training.supersonic)
    surrogates.transonic   =  build_surrogate(aerodynamics, training.transonic)
    return

def build_surrogate(aerodynamics, training):
    
    # unpack data
    surrogates     = Data()
    mach_data      = training.Mach
    vehicle        = aerodynamics.vehicle
    AoA_data       = aerodynamics.training.angle_of_attack           
    Beta_data      = aerodynamics.training.sideslip_angle  
    u_data         = aerodynamics.training.u
    v_data         = aerodynamics.training.v
    w_data         = aerodynamics.training.w
    p_data         = aerodynamics.training.roll_rate
    q_data         = aerodynamics.training.pitch_rate
    r_data         = aerodynamics.training.yaw_rate
    aileron_data   = aerodynamics.training.aileron_deflection           
    elevator_data  = aerodynamics.training.elevator_deflection          
    rudder_data    = aerodynamics.training.rudder_deflection            
    flap_data      = aerodynamics.training.flap_deflection              
    slat_data      = aerodynamics.training.slat_deflection
    
    surrogates.Clift_wing_alpha = Data()
    surrogates.Cdrag_wing_alpha = Data() 
    for wing in  vehicle.wings: 
        surrogates.Clift_wing_alpha[wing.tag] = RegularGridInterpolator((AoA_data ,mach_data),training.Clift_wing_alpha[wing.tag],method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.Cdrag_wing_alpha[wing.tag] = RegularGridInterpolator((AoA_data ,mach_data),training.Cdrag_wing_alpha[wing.tag],method = 'linear',   bounds_error=False, fill_value=None) 
     
    # Pack the outputs     
    surrogates.Clift_alpha       = RegularGridInterpolator((AoA_data ,mach_data),training.Clift_alpha        ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.Clift_beta        = RegularGridInterpolator((Beta_data,mach_data),training.Clift_beta         ,method = 'linear',   bounds_error=False, fill_value=None) 
    surrogates.Clift_u           = RegularGridInterpolator((u_data,mach_data),training.Clift_u               ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.Clift_v           = RegularGridInterpolator((v_data,mach_data),training.Clift_v               ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.Clift_w           = RegularGridInterpolator((w_data,mach_data),training.Clift_w               ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.Clift_p           = RegularGridInterpolator((p_data,mach_data),training.Clift_p               ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.Clift_q           = RegularGridInterpolator((q_data,mach_data),training.Clift_q               ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.Clift_r           = RegularGridInterpolator((r_data,mach_data),training.Clift_r               ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.Cdrag_alpha       = RegularGridInterpolator((AoA_data,mach_data),training.Cdrag_alpha         ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.Cdrag_beta        = RegularGridInterpolator((Beta_data,mach_data),training.Cdrag_beta         ,method = 'linear',   bounds_error=False, fill_value=None)
     
              
    surrogates.Cdrag_u        = RegularGridInterpolator((u_data,mach_data),training.Cdrag_u        ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.Cdrag_v        = RegularGridInterpolator((v_data,mach_data),training.Cdrag_v        ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.Cdrag_w        = RegularGridInterpolator((w_data,mach_data),training.Cdrag_w        ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.Cdrag_p        = RegularGridInterpolator((p_data,mach_data),training.Cdrag_p        ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.Cdrag_q        = RegularGridInterpolator((q_data,mach_data),training.Cdrag_q        ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.Cdrag_r        = RegularGridInterpolator((r_data,mach_data),training.Cdrag_r        ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CX_alpha       = RegularGridInterpolator((AoA_data,mach_data),training.CX_alpha     ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CX_beta        = RegularGridInterpolator((Beta_data,mach_data),training.CX_beta     ,method = 'linear',   bounds_error=False, fill_value=None) 
     
                    
    surrogates.CX_u           = RegularGridInterpolator((u_data,mach_data),training.CX_u           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CX_v           = RegularGridInterpolator((v_data,mach_data),training.CX_v           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CX_w           = RegularGridInterpolator((w_data,mach_data),training.CX_w           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CX_p           = RegularGridInterpolator((p_data,mach_data),training.CX_p           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CX_q           = RegularGridInterpolator((q_data,mach_data),training.CX_q           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CX_r           = RegularGridInterpolator((r_data,mach_data),training.CX_r           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CY_alpha       = RegularGridInterpolator((AoA_data,mach_data),training.CY_alpha     ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CY_beta        = RegularGridInterpolator((Beta_data,mach_data),training.CY_beta     ,method = 'linear',   bounds_error=False, fill_value=None)
     
              
    surrogates.CY_u           = RegularGridInterpolator((u_data,mach_data),training.CY_u           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CY_v           = RegularGridInterpolator((v_data,mach_data),training.CY_v           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CY_w           = RegularGridInterpolator((w_data,mach_data),training.CY_w           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CY_p           = RegularGridInterpolator((p_data,mach_data),training.CY_p           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CY_q           = RegularGridInterpolator((q_data,mach_data),training.CY_q           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CY_r           = RegularGridInterpolator((r_data,mach_data),training.CY_r           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CZ_alpha       = RegularGridInterpolator((AoA_data,mach_data),training.CZ_alpha     ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CZ_beta        = RegularGridInterpolator((Beta_data,mach_data),training.CZ_beta     ,method = 'linear',   bounds_error=False, fill_value=None)
     
                
    surrogates.CZ_u           = RegularGridInterpolator((u_data,mach_data),training.CZ_u           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CZ_v           = RegularGridInterpolator((v_data,mach_data),training.CZ_v           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CZ_w           = RegularGridInterpolator((w_data,mach_data),training.CZ_w           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CZ_p           = RegularGridInterpolator((p_data,mach_data),training.CZ_p           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CZ_q           = RegularGridInterpolator((q_data,mach_data),training.CZ_q           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CZ_r           = RegularGridInterpolator((r_data,mach_data),training.CZ_r           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CL_alpha       = RegularGridInterpolator((AoA_data,mach_data),training.CL_alpha     ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CL_beta        = RegularGridInterpolator((Beta_data,mach_data),training.CL_beta     ,method = 'linear',   bounds_error=False, fill_value=None)
     
               
    surrogates.CL_u           = RegularGridInterpolator((u_data,mach_data),training.CL_u           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CL_v           = RegularGridInterpolator((v_data,mach_data),training.CL_v           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CL_w           = RegularGridInterpolator((w_data,mach_data),training.CL_w           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CL_p           = RegularGridInterpolator((p_data,mach_data),training.CL_p           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CL_q           = RegularGridInterpolator((q_data,mach_data),training.CL_q           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CL_r           = RegularGridInterpolator((r_data,mach_data),training.CL_r           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CM_alpha       = RegularGridInterpolator((AoA_data,mach_data),training.CM_alpha     ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CM_beta        = RegularGridInterpolator((Beta_data,mach_data),training.CM_beta     ,method = 'linear',   bounds_error=False, fill_value=None)
    
    surrogates.CM_u           = RegularGridInterpolator((u_data,mach_data),training.CM_u           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CM_v           = RegularGridInterpolator((v_data,mach_data),training.CM_v           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CM_w           = RegularGridInterpolator((w_data,mach_data),training.CM_w           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CM_p           = RegularGridInterpolator((p_data,mach_data),training.CM_p           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CM_q           = RegularGridInterpolator((q_data,mach_data),training.CM_q           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CM_r           = RegularGridInterpolator((r_data,mach_data),training.CM_r           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CN_alpha       = RegularGridInterpolator((AoA_data,mach_data),training.CN_alpha     ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CN_beta        = RegularGridInterpolator((Beta_data,mach_data),training.CN_beta     ,method = 'linear',   bounds_error=False, fill_value=None)
             
    surrogates.CN_u           = RegularGridInterpolator((u_data,mach_data),training.CN_u           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CN_v           = RegularGridInterpolator((v_data,mach_data),training.CN_v           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CN_w           = RegularGridInterpolator((w_data,mach_data),training.CN_w           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CN_p           = RegularGridInterpolator((p_data,mach_data),training.CN_p           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CN_q           = RegularGridInterpolator((q_data,mach_data),training.CN_q           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.CN_r           = RegularGridInterpolator((r_data,mach_data),training.CN_r           ,method = 'linear',   bounds_error=False, fill_value=None)
    
    surrogates.dClift_dalpha    = interpolate.interp1d(mach_data,training.dClift_dalpha    ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dClift_dbeta     = interpolate.interp1d(mach_data,training.dClift_dbeta     ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate") 
    surrogates.dClift_du        = interpolate.interp1d(mach_data,training.dClift_du        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dClift_dv        = interpolate.interp1d(mach_data,training.dClift_dv        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dClift_dw        = interpolate.interp1d(mach_data,training.dClift_dw        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dClift_dp        = interpolate.interp1d(mach_data,training.dClift_dp        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dClift_dq        = interpolate.interp1d(mach_data,training.dClift_dq        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dClift_dr        = interpolate.interp1d(mach_data,training.dClift_dr        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCdrag_dalpha    = interpolate.interp1d(mach_data,training.dCdrag_dalpha    ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCdrag_dbeta     = interpolate.interp1d(mach_data,training.dCdrag_dbeta     ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")   
    surrogates.dCdrag_du        = interpolate.interp1d(mach_data,training.dCdrag_du        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCdrag_dv        = interpolate.interp1d(mach_data,training.dCdrag_dv        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCdrag_dw        = interpolate.interp1d(mach_data,training.dCdrag_dw        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCdrag_dp        = interpolate.interp1d(mach_data,training.dCdrag_dp        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCdrag_dq        = interpolate.interp1d(mach_data,training.dCdrag_dq        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCdrag_dr        = interpolate.interp1d(mach_data,training.dCdrag_dr        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCX_dalpha       = interpolate.interp1d(mach_data,training.dCX_dalpha       ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCX_dbeta        = interpolate.interp1d(mach_data,training.dCX_dbeta        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")  
    surrogates.dCX_du           = interpolate.interp1d(mach_data,training.dCX_du           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCX_dv           = interpolate.interp1d(mach_data,training.dCX_dv           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCX_dw           = interpolate.interp1d(mach_data,training.dCX_dw           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCX_dp           = interpolate.interp1d(mach_data,training.dCX_dp           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCX_dq           = interpolate.interp1d(mach_data,training.dCX_dq           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCX_dr           = interpolate.interp1d(mach_data,training.dCX_dr           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCY_dalpha       = interpolate.interp1d(mach_data,training.dCY_dalpha       ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCY_dbeta        = interpolate.interp1d(mach_data,training.dCY_dbeta        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")  
    surrogates.dCY_du           = interpolate.interp1d(mach_data,training.dCY_du           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCY_dv           = interpolate.interp1d(mach_data,training.dCY_dv           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCY_dw           = interpolate.interp1d(mach_data,training.dCY_dw           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCY_dp           = interpolate.interp1d(mach_data,training.dCY_dp           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCY_dq           = interpolate.interp1d(mach_data,training.dCY_dq           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCY_dr           = interpolate.interp1d(mach_data,training.dCY_dr           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCZ_dalpha       = interpolate.interp1d(mach_data,training.dCZ_dalpha       ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCZ_dbeta        = interpolate.interp1d(mach_data,training.dCZ_dbeta        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")  
    surrogates.dCZ_du           = interpolate.interp1d(mach_data,training.dCZ_du           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCZ_dv           = interpolate.interp1d(mach_data,training.dCZ_dv           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCZ_dw           = interpolate.interp1d(mach_data,training.dCZ_dw           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCZ_dp           = interpolate.interp1d(mach_data,training.dCZ_dp           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCZ_dq           = interpolate.interp1d(mach_data,training.dCZ_dq           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCZ_dr           = interpolate.interp1d(mach_data,training.dCZ_dr           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCL_dalpha       = interpolate.interp1d(mach_data,training.dCL_dalpha       ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCL_dbeta        = interpolate.interp1d(mach_data,training.dCL_dbeta        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")  
    surrogates.dCL_du           = interpolate.interp1d(mach_data,training.dCL_du           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCL_dv           = interpolate.interp1d(mach_data,training.dCL_dv           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCL_dw           = interpolate.interp1d(mach_data,training.dCL_dw           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCL_dp           = interpolate.interp1d(mach_data,training.dCL_dp           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCL_dq           = interpolate.interp1d(mach_data,training.dCL_dq           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCL_dr           = interpolate.interp1d(mach_data,training.dCL_dr           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCM_dalpha       = interpolate.interp1d(mach_data,training.dCM_dalpha       ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCM_dbeta        = interpolate.interp1d(mach_data,training.dCM_dbeta        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")  
    surrogates.dCM_du           = interpolate.interp1d(mach_data,training.dCM_du           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCM_dv           = interpolate.interp1d(mach_data,training.dCM_dv           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCM_dw           = interpolate.interp1d(mach_data,training.dCM_dw           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCM_dp           = interpolate.interp1d(mach_data,training.dCM_dp           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCM_dq           = interpolate.interp1d(mach_data,training.dCM_dq           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCM_dr           = interpolate.interp1d(mach_data,training.dCM_dr           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCN_dalpha       = interpolate.interp1d(mach_data,training.dCN_dalpha       ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCN_dbeta        = interpolate.interp1d(mach_data,training.dCN_dbeta        ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")  
    surrogates.dCN_du           = interpolate.interp1d(mach_data,training.dCN_du           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCN_dv           = interpolate.interp1d(mach_data,training.dCN_dv           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCN_dw           = interpolate.interp1d(mach_data,training.dCN_dw           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCN_dp           = interpolate.interp1d(mach_data,training.dCN_dp           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCN_dq           = interpolate.interp1d(mach_data,training.dCN_dq           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")      
    surrogates.dCN_dr           = interpolate.interp1d(mach_data,training.dCN_dr           ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")
   

    if aerodynamics.aileron_flag: 
        surrogates.Clift_delta_a    = RegularGridInterpolator((aileron_data,mach_data),training.Clift_delta_a  ,method = 'linear',   bounds_error=False, fill_value=None)
        surrogates.Cdrag_delta_a    = RegularGridInterpolator((aileron_data,mach_data),training.Cdrag_delta_a  ,method = 'linear',   bounds_error=False, fill_value=None)
        surrogates.CX_delta_a       = RegularGridInterpolator((aileron_data,mach_data),training.CX_delta_a  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CY_delta_a       = RegularGridInterpolator((aileron_data,mach_data),training.CY_delta_a  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CZ_delta_a       = RegularGridInterpolator((aileron_data,mach_data),training.CZ_delta_a  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CL_delta_a       = RegularGridInterpolator((aileron_data,mach_data),training.CL_delta_a  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CM_delta_a       = RegularGridInterpolator((aileron_data,mach_data),training.CM_delta_a  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CN_delta_a       = RegularGridInterpolator((aileron_data,mach_data),training.CN_delta_a  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.dClift_ddelta_a  = interpolate.interp1d(mach_data,training.dClift_ddelta_a     , kind = 'linear',   bounds_error=False, fill_value="extrapolate")
        surrogates.dCdrag_ddelta_a  = interpolate.interp1d(mach_data,training.dCdrag_ddelta_a     , kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCX_ddelta_a     = interpolate.interp1d(mach_data,training.dCX_ddelta_a        , kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCY_ddelta_a     = interpolate.interp1d(mach_data,training.dCY_ddelta_a        , kind = 'linear',   bounds_error=False, fill_value="extrapolate")  
        surrogates.dCZ_ddelta_a     = interpolate.interp1d(mach_data,training.dCZ_ddelta_a        , kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCL_ddelta_a     = interpolate.interp1d(mach_data,training.dCL_ddelta_a        , kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCM_ddelta_a     = interpolate.interp1d(mach_data,training.dCM_ddelta_a        , kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCN_ddelta_a     = interpolate.interp1d(mach_data,training.dCN_ddelta_a        , kind = 'linear',   bounds_error=False, fill_value="extrapolate")             
    
    if aerodynamics.elevator_flag: 
        surrogates.Clift_delta_e    = RegularGridInterpolator((elevator_data,mach_data),training.Clift_delta_e  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.Cdrag_delta_e    = RegularGridInterpolator((elevator_data,mach_data),training.Cdrag_delta_e  ,method = 'linear',   bounds_error=False, fill_value=None)
        surrogates.CX_delta_e       = RegularGridInterpolator((elevator_data,mach_data),training.CX_delta_e  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CY_delta_e       = RegularGridInterpolator((elevator_data,mach_data),training.CY_delta_e  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CZ_delta_e       = RegularGridInterpolator((elevator_data,mach_data),training.CZ_delta_e  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CL_delta_e       = RegularGridInterpolator((elevator_data,mach_data),training.CL_delta_e  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CM_delta_e       = RegularGridInterpolator((elevator_data,mach_data),training.CM_delta_e  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CN_delta_e       = RegularGridInterpolator((elevator_data,mach_data),training.CN_delta_e  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.dClift_ddelta_e  = interpolate.interp1d(mach_data,training.dClift_ddelta_e  ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")
        surrogates.dCdrag_ddelta_e  = interpolate.interp1d(mach_data,training.dCdrag_ddelta_e  ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate") 
        surrogates.dCX_ddelta_e     = interpolate.interp1d(mach_data,training.dCX_ddelta_e     ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate") 
        surrogates.dCY_ddelta_e     = interpolate.interp1d(mach_data,training.dCY_ddelta_e     ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")  
        surrogates.dCZ_ddelta_e     = interpolate.interp1d(mach_data,training.dCZ_ddelta_e     ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate") 
        surrogates.dCL_ddelta_e     = interpolate.interp1d(mach_data,training.dCL_ddelta_e     ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate") 
        surrogates.dCM_ddelta_e     = interpolate.interp1d(mach_data,training.dCM_ddelta_e     ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate") 
        surrogates.dCN_ddelta_e     = interpolate.interp1d(mach_data,training.dCN_ddelta_e     ,kind = 'linear',   bounds_error=False, fill_value= "extrapolate")   
    
    
    if aerodynamics.rudder_flag: 
        surrogates.Clift_delta_r    = RegularGridInterpolator((rudder_data,mach_data),training.Clift_delta_r  ,method = 'linear',   bounds_error=False, fill_value=None)  
        surrogates.Cdrag_delta_r    = RegularGridInterpolator((rudder_data,mach_data),training.Cdrag_delta_r  ,method = 'linear',   bounds_error=False, fill_value=None)
        surrogates.CX_delta_r       = RegularGridInterpolator((rudder_data,mach_data),training.CX_delta_r  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CY_delta_r       = RegularGridInterpolator((rudder_data,mach_data),training.CY_delta_r  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CZ_delta_r       = RegularGridInterpolator((rudder_data,mach_data),training.CZ_delta_r  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CL_delta_r       = RegularGridInterpolator((rudder_data,mach_data),training.CL_delta_r  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CM_delta_r       = RegularGridInterpolator((rudder_data,mach_data),training.CM_delta_r  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CN_delta_r       = RegularGridInterpolator((rudder_data,mach_data),training.CN_delta_r  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.dClift_ddelta_r  = interpolate.interp1d(mach_data,training.dClift_ddelta_r    ,kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCdrag_ddelta_r  = interpolate.interp1d(mach_data,training.dCdrag_ddelta_r    ,kind = 'linear',   bounds_error=False, fill_value="extrapolate")  
        surrogates.dCX_ddelta_r     = interpolate.interp1d(mach_data,training.dCX_ddelta_r       ,kind = 'linear',   bounds_error=False, fill_value="extrapolate")  
        surrogates.dCY_ddelta_r     = interpolate.interp1d(mach_data,training.dCY_ddelta_r       ,kind = 'linear',   bounds_error=False, fill_value="extrapolate")   
        surrogates.dCZ_ddelta_r     = interpolate.interp1d(mach_data,training.dCZ_ddelta_r       ,kind = 'linear',   bounds_error=False, fill_value="extrapolate")  
        surrogates.dCL_ddelta_r     = interpolate.interp1d(mach_data,training.dCL_ddelta_r       ,kind = 'linear',   bounds_error=False, fill_value="extrapolate")  
        surrogates.dCM_ddelta_r     = interpolate.interp1d(mach_data,training.dCM_ddelta_r       ,kind = 'linear',   bounds_error=False, fill_value="extrapolate")  
        surrogates.dCN_ddelta_r     = interpolate.interp1d(mach_data,training.dCN_ddelta_r       ,kind = 'linear',   bounds_error=False, fill_value="extrapolate")    
    
    if aerodynamics.flap_flag:
        surrogates.Clift_delta_f    = RegularGridInterpolator((flap_data,mach_data),training.Clift_delta_f  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.Cdrag_delta_f    = RegularGridInterpolator((flap_data,mach_data),training.Cdrag_delta_f  ,method = 'linear',   bounds_error=False, fill_value=None)
        surrogates.CX_delta_f       = RegularGridInterpolator((flap_data,mach_data),training.CX_delta_f  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CY_delta_f       = RegularGridInterpolator((flap_data,mach_data),training.CY_delta_f  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CZ_delta_f       = RegularGridInterpolator((flap_data,mach_data),training.CZ_delta_f  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CL_delta_f       = RegularGridInterpolator((flap_data,mach_data),training.CL_delta_f  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CM_delta_f       = RegularGridInterpolator((flap_data,mach_data),training.CM_delta_f  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CN_delta_f       = RegularGridInterpolator((flap_data,mach_data),training.CN_delta_f  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.dClift_ddelta_f  = interpolate.interp1d(mach_data,training.dClift_ddelta_f  ,kind = 'linear',   bounds_error=False, fill_value="extrapolate")
        surrogates.dCdrag_ddelta_f  = interpolate.interp1d(mach_data,training.dCdrag_ddelta_f  ,kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCX_ddelta_f     = interpolate.interp1d(mach_data,training.dCX_ddelta_f     ,kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCY_ddelta_f     = interpolate.interp1d(mach_data,training.dCY_ddelta_f     ,kind = 'linear',   bounds_error=False, fill_value="extrapolate")  
        surrogates.dCZ_ddelta_f     = interpolate.interp1d(mach_data,training.dCZ_ddelta_f     ,kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCL_ddelta_f     = interpolate.interp1d(mach_data,training.dCL_ddelta_f     ,kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCM_ddelta_f     = interpolate.interp1d(mach_data,training.dCM_ddelta_f     ,kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCN_ddelta_f     = interpolate.interp1d(mach_data,training.dCN_ddelta_f     ,kind = 'linear',   bounds_error=False, fill_value="extrapolate")   
    
    if aerodynamics.slat_flag: 
        surrogates.Clift_delta_s    = RegularGridInterpolator((slat_data,mach_data),training.Clift_delta_s  ,method = 'linear',   bounds_error=False, fill_value=None)         
        surrogates.Cdrag_delta_s    = RegularGridInterpolator((slat_data,mach_data),training.Cdrag_delta_s  ,method = 'linear',   bounds_error=False, fill_value=None)
        surrogates.CX_delta_s       = RegularGridInterpolator((slat_data,mach_data),training.CX_delta_s  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CY_delta_s       = RegularGridInterpolator((slat_data,mach_data),training.CY_delta_s  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CZ_delta_s       = RegularGridInterpolator((slat_data,mach_data),training.CZ_delta_s  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CL_delta_s       = RegularGridInterpolator((slat_data,mach_data),training.CL_delta_s  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CM_delta_s       = RegularGridInterpolator((slat_data,mach_data),training.CM_delta_s  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.CN_delta_s       = RegularGridInterpolator((slat_data,mach_data),training.CN_delta_s  ,method = 'linear',   bounds_error=False, fill_value=None) 
        surrogates.dClift_ddelta_s  = interpolate.interp1d(mach_data,training.dClift_ddelta_s  ,kind = 'linear',   bounds_error=False, fill_value="extrapolate")  
        surrogates.dCdrag_ddelta_s  = interpolate.interp1d(mach_data,training.dCdrag_ddelta_s  ,kind = 'linear',   bounds_error=False, fill_value="extrapolate")   
        surrogates.dCX_ddelta_s     = interpolate.interp1d(mach_data,training.dCX_ddelta_s     ,kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCY_ddelta_s     = interpolate.interp1d(mach_data,training.dCY_ddelta_s     ,kind = 'linear',   bounds_error=False, fill_value="extrapolate")  
        surrogates.dCZ_ddelta_s     = interpolate.interp1d(mach_data,training.dCZ_ddelta_s     ,kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCL_ddelta_s     = interpolate.interp1d(mach_data,training.dCL_ddelta_s     ,kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCM_ddelta_s     = interpolate.interp1d(mach_data,training.dCM_ddelta_s     ,kind = 'linear',   bounds_error=False, fill_value="extrapolate") 
        surrogates.dCN_ddelta_s     = interpolate.interp1d(mach_data,training.dCN_ddelta_s     ,kind = 'linear',   bounds_error=False, fill_value="extrapolate")   
   
    return surrogates
 


# RCAIDE/Methods/Aerodynamics/Common/Lift/BET_calculations.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
 
from RCAIDE.Framework.Core import interp2d 

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  compute_airfoil_aerodynamics
# ----------------------------------------------------------------------------------------------------------------------   
def compute_airfoil_aerodynamics(beta,c,r,R,B,Wa,Wt,a,nu,airfoils,a_loc,ctrl_pts,Nr,Na,tc,use_2d_analysis):
    """
    Cl, Cdval = compute_airfoil_aerodynamics( beta,c,r,R,B,
                                              Wa,Wt,a,nu,
                                              airfoils,a_loc
                                              ctrl_pts,Nr,Na,tc,use_2d_analysis )

    Computes the aerodynamic forces at sectional blade locations. If airfoil
    geometry and locations are specified, the forces are computed using the
    airfoil polar lift and drag surrogates, accounting for the local Reynolds
    number and local angle of attack.

    If the airfoils are not specified, an approximation is used.

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
       beta                       blade twist distribution                        [-]
       c                          chord distribution                              [-]
       r                          radius distribution                             [-]
       R                          tip radius                                      [-]
       B                          number of rotor blades                          [-]

       Wa                         axial velocity                                  [-]
       Wt                         tangential velocity                             [-]
       a                          speed of sound                                  [-]
       nu                         viscosity                                       [-]
       airfoil_data               Data structure of airfoil polar information     [-]
       ctrl_pts                   Number of control points                        [-]
       Nr                         Number of radial blade sections                 [-]
       Na                         Number of azimuthal blade stations              [-]
       tc                         Thickness to chord                              [-]
       use_2d_analysis            Specifies 2d disc vs. 1d single angle analysis  [Boolean]

    Outputs:
       Cl                       Lift Coefficients                         [-]
       Cdval                    Drag Coefficients  (before scaling)       [-]
       alpha                    section local angle of attack             [rad]

    """ 
    alpha    = beta - np.arctan2(Wa,Wt)
    W        = (Wa*Wa + Wt*Wt)**0.5
    Ma       = W/a
    Re       = (W*c)/nu

    # If rotor airfoils are defined, use airfoil surrogate
    a_loc = np.array(a_loc)
    if np.any(a_loc) != None:  
        # Compute blade Cl and Cd distribution from the airfoil data 
        if use_2d_analysis:
            # return the 2D Cl and CDval of shape (ctrl_pts, Nr, Na)
            Cl         = np.zeros((ctrl_pts,Nr,Na))
            Cdval      = np.zeros((ctrl_pts,Nr,Na))
            
            for jj,airfoil in enumerate(airfoils):
                pd                   = airfoil.polars
                Cl_af                = interp2d(Re,alpha,pd.reynolds_numbers, pd.angle_of_attacks, pd.lift_coefficients) 
                Cdval_af             = interp2d(Re,alpha,pd.reynolds_numbers, pd.angle_of_attacks, pd.drag_coefficients)
                locs                 = np.where(np.array(a_loc) == jj )
                Cl[:,locs,:]         = Cl_af[:,locs,:]
                Cdval[:,locs,:]      = Cdval_af[:,locs,:]
            alpha_disc           = alpha
            Re_disc              = Re
        else:
            # return the 1D Cl and CDval of shape (ctrl_pts, Nr)
            Cl         = np.zeros((ctrl_pts,Nr))
            Cdval      = np.zeros((ctrl_pts,Nr))             

            for jj,airfoil in enumerate(airfoils):
                pd                   = airfoil.polars
                Cl_af                = interp2d(Re,alpha,pd.reynolds_numbers, pd.angle_of_attacks, pd.lift_coefficients)
                Cdval_af             = interp2d(Re,alpha,pd.reynolds_numbers, pd.angle_of_attacks, pd.drag_coefficients)
                locs                 = np.where(np.array(a_loc) == jj )
                Cl[:,locs]           = Cl_af[:,locs]
                Cdval[:,locs]        = Cdval_af[:,locs] 
            alpha_disc = np.tile(alpha[:,:, None], (1, 1, Na)) 
            Re_disc    = np.tile(Re[:,:, None], (1, 1, Na))  

    else:
        # Estimate Cl max
        tc_1 = tc*100
        Cl_max_ref = -0.0009*tc_1**3 + 0.0217*tc_1**2 - 0.0442*tc_1 + 0.7005
        Cl_max_ref[Cl_max_ref<0.7] = 0.7
        Re_ref     = 9.*10**6
        Cl1maxp    = Cl_max_ref * ( Re / Re_ref ) **0.1

        # If not airfoil polar provided, use 2*pi as lift curve slope
        Cl = 2.*np.pi*alpha

        # By 90 deg, it's totally stalled.
        Cl[Cl>Cl1maxp]  = Cl1maxp[Cl>Cl1maxp] # This line of code is what changed the regression testing
        Cl[alpha>=np.pi/2] = 0.

        # Scale for Mach, this is Karmen_Tsien
        KT_cond = np.logical_and((Ma[:,:]<1.),(Cl>0))
        Cl[KT_cond] = Cl[KT_cond]/((1-Ma[KT_cond]*Ma[KT_cond])**0.5+((Ma[KT_cond]*Ma[KT_cond])/(1+(1-Ma[KT_cond]*Ma[KT_cond])**0.5))*Cl[KT_cond]/2)

        # If the blade segments are supersonic, don't scale
        Cl[Ma[:,:]>=1.] = Cl[Ma[:,:]>=1.]

        #This is an atrocious fit of DAE51 data at RE=50k for Cd
        Cdval = (0.108*(Cl*Cl*Cl*Cl)-0.2612*(Cl*Cl*Cl)+0.181*(Cl*Cl)-0.0139*Cl+0.0278)*((50000./Re)**0.2)
        Cdval[alpha>=np.pi/2] = 2.
        
        alpha_disc = np.tile(alpha[:,:, None], (1, 1, Nr)) 
        Re_disc    = np.tile(Re[:,:, None], (1, 1, Nr))  

    # prevent zero Cl to keep Cd/Cl from breaking in BET
    Cl[Cl==0] = 1e-6

    return Cl, Cdval, alpha, alpha_disc,Ma,W,Re,Re_disc

# ----------------------------------------------------------------------------------------------------------------------
#  compute_inflow_and_tip_loss
# ----------------------------------------------------------------------------------------------------------------------    
def compute_inflow_and_tip_loss(r,R,Wa,Wt,B,et1=1,et2=1,et3=1):
    """
    Computes the inflow, lamdaw, and the tip loss factor, F.

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
       r          radius distribution                                              [m]
       R          tip radius                                                       [m]
       Wa         axial velocity                                                   [m/s]
       Wt         tangential velocity                                              [m/s]
       B          number of rotor blades                                           [-]
       et1        tuning parameter for tip loss function 
       et2        tuning parameter for tip loss function 
       et3        tuning parameter for tip loss function 
       
    Outputs:               
       lamdaw     inflow ratio                                                     [-]
       F          tip loss factor                                                  [-]
       piece      output of a step in tip loss calculation (needed for residual)   [-]
    """
    lamdaw             = r*Wa/(R*Wt)
    lamdaw[lamdaw<=0.] = 1e-12

    tipfactor = B/2.0*(  (R/r)**et1 - 1  )**et2/lamdaw**et3 

    piece = np.exp(-tipfactor)
    Ftip  = 2.*np.arccos(piece)/np.pi  

    return lamdaw, Ftip, piece


# RCAIDE/Library/Methods/Propulsors/Converters/Engine/append_engine_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_engine_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_engine_conditions(engine,segment,propulsor_conditions): 
    propulsor_conditions[engine.tag]                      = Conditions() 
    propulsor_conditions[engine.tag].inputs               = Conditions()
    propulsor_conditions[engine.tag].outputs              = Conditions()      
    return 



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/ccompute_propulsion_system_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import  RCAIDE 
from RCAIDE.Framework.Core    import Units ,  Data

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Propulsion Systems Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_propulsion_system_weight(vehicle,ref_propulsor):
    """ Calculate the weight of propulsion system, including:
        - dry engine weight
        - fuel system weight
        - thurst reversers weight
        - electrical system weight
        - starter engine weight
        - nacelle weight
        - cargo containers

        Assumptions:
            1) Rated thrust per scaled engine and rated thurst for baseline are the same
            2) Engine weight scaling parameter is 1.15
            3) Enginge inlet weight scaling exponent is 1
            4) Baseline inlet weight is 0 lbs as in example files FLOPS
            5) Baseline nozzle weight is 0 lbs as in example files FLOPS

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.design_mach_number: design mach number for cruise flight
                -.mass_properties.max_zero_fuel: zero fuel weight               [kg]
                -.systems.accessories: type of aircraft (short-range, commuter
                                                        medium-range, long-range,
                                                        sst, cargo)
            nacelle - data dictionary with propulsion system properties 
                -.diameter: diameter of nacelle                                 [meters]
                -.length: length of complete engine assembly                    [meters]
            ref_propulsor.
                -.sealevel_static_thrust: thrust at sea level                   [N]


        Outputs:
            output - data dictionary with weights                               [kilograms]
                    - output.W_prop: total propulsive system weight
                    - output.W_thrust_reverser: thurst reverser weight
                    - output.starter: starter engine weight
                    - output.W_engine_controls: engine controls weight
                    - output.fuel_system: fuel system weight
                    - output.nacelle: nacelle weight
                    - output.W_engine: dry engine weight

        Properties Used:
            N/A
    """
     
    NENG =  0 
    number_of_tanks =  0
    for network in  vehicle.networks:
        for propulsor in network.propulsors:
            if isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbofan) or  isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbojet):
                ref_propulsor = propulsor  
                NENG  += 1 
            if 'nacelle' in propulsor:
                ref_nacelle =  propulsor.nacelle   
        for fuel_line in network.fuel_lines:
            for fuel_tank in fuel_line.fuel_tanks:
                number_of_tanks +=  1
                  
     
    WNAC            = nacelle_FLOPS(ref_propulsor,ref_nacelle,NENG ) 
    WFSYS           = fuel_system_FLOPS(vehicle, NENG)
    WENG            = compute_engine_weight(vehicle,ref_propulsor)
    WEC, WSTART     = misc_engine_FLOPS(vehicle,ref_propulsor,ref_nacelle,NENG)
    WTHR            = thrust_reverser_FLOPS(ref_propulsor,NENG)
    WPRO            = NENG * WENG + WFSYS + WEC + WSTART + WTHR + WNAC

    output                      = Data()
    output.W_prop               = WPRO
    output.W_thrust_reverser    = WTHR
    output.W_starter            = WSTART
    output.W_engine_controls    = WEC
    output.W_fuel_system        = WFSYS
    output.W_nacelle            = WNAC
    output.W_engine             = WENG * NENG
    output.number_of_engines    = NENG 
    output.number_of_fuel_tanks = number_of_tanks  
    return output


def nacelle_FLOPS(ref_propulsor,ref_nacelle,NENG):
    """ Calculates the nacelle weight based on the FLOPS method
    
        Assumptions:
            1) All nacelles are identical
            2) The number of nacelles is the same as the number of engines 

        Source:
            Aircraft Design: A Conceptual Approach

        Inputs:
            ref_propulsor    - data dictionary for the specific network that is being estimated [dimensionless]
                -.number_of_engines: number of engines
                -.engine_lenght: total length of engine                                  [m]
                -.sealevel_static_thrust: sealevel static thrust of engine               [N]
            nacelle.             
                -.diameter: diameter of nacelle                                          [m]
            WENG    - dry engine weight                                                  [kg]
             
             
        Outputs:             
            WNAC: nacelle weight                                                         [kg]

        Properties Used:
            N/A
    """ 
    TNAC   = NENG + 1. / 2 * (NENG - 2 * np.floor(NENG / 2.))
    DNAC   = ref_nacelle.diameter / Units.ft
    XNAC   = ref_nacelle.length / Units.ft
    FTHRST = ref_propulsor.sealevel_static_thrust * 1 / Units.lbf
    WNAC   = 0.25 * TNAC * DNAC * XNAC * FTHRST ** 0.36
    return WNAC * Units.lbs


def thrust_reverser_FLOPS(ref_propulsor,NENG):
    """ Calculates the weight of the thrust reversers of the aircraft
    
        Assumptions:

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            ref_propulsor    - data dictionary for the specific network that is being estimated [dimensionless]
                -.number_of_engines: number of engines
                -.sealevel_static_thrust: sealevel static thrust of engine  [N]

        Outputs:
            WTHR: Thrust reversers weight                                   [kg]

        Properties Used:
            N/A
    """ 
    TNAC = NENG + 1. / 2 * (NENG - 2 * np.floor(NENG / 2.))
    THRUST = ref_propulsor.sealevel_static_thrust * 1 / Units.lbf
    WTHR = 0.034 * THRUST * TNAC
    return WTHR * Units.lbs


def misc_engine_FLOPS(vehicle,ref_propulsor,ref_nacelle,NENG ):
    """ Calculates the miscellaneous engine weight based on the FLOPS method, electrical control system weight
        and starter engine weight
        
        Assumptions:
            1) All nacelles are identical
            2) The number of nacelles is the same as the number of engines 

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            vehicle - data dictionary with vehicle properties                            [dimensionless]
                 -.design_mach_number: design mach number
            ref_propulsor    - data dictionary for the specific network that is being estimated [dimensionless]
                -.number_of_engines: number of engines
                -.sealevel_static_thrust: sealevel static thrust of engine               [N]
            nacelle              
                -.diameter: diameter of nacelle                                          [m]
              
        Outputs:              
            WEC: electrical engine control system weight                                 [kg]
            WSTART: starter engine weight                                                [kg]

        Properties Used:
            N/A
    """ 
    THRUST  = ref_propulsor.sealevel_static_thrust * 1 / Units.lbf
    WEC     = 0.26 * NENG * THRUST ** 0.5
    FNAC    = ref_nacelle.diameter / Units.ft
    VMAX    = vehicle.flight_envelope.design_mach_number
    WSTART  = 11.0 * NENG * VMAX ** 0.32 * FNAC ** 1.6
    return WEC * Units.lbs, WSTART * Units.lbs


def fuel_system_FLOPS(vehicle, NENG):
    """ Calculates the weight of the fuel system based on the FLOPS method
        Assumptions:

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.design_mach_number: design mach number
                -.mass_properties.max_zero_fuel: maximum zero fuel weight   [kg]

        Outputs:
            WFSYS: Fuel system weight                                       [kg]

        Properties Used:
            N/A
    """
    VMAX = vehicle.flight_envelope.design_mach_number
    FMXTOT = vehicle.mass_properties.max_zero_fuel / Units.lbs
    WFSYS = 1.07 * FMXTOT ** 0.58 * NENG ** 0.43 * VMAX ** 0.34
    return WFSYS * Units.lbs


def compute_engine_weight(vehicle, ref_propulsor):
    """ Calculates the dry engine weight based on the FLOPS method
        Assumptions:
            Rated thrust per scaled engine and rated thurst for baseline are the same
            Engine weight scaling parameter is 1.15
            Enginge inlet weight scaling exponent is 1
            Baseline inlet weight is 0 lbs as in example files FLOPS
            Baseline nozzle weight is 0 lbs as in example files FLOPS

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.systems.accessories: type of aircraft (short-range, commuter
                                                        medium-range, long-range,
                                                        sst, cargo)
            ref_propulsor    - data dictionary for the specific network that is being estimated [dimensionless]
                -.sealevel_static_thrust: sealevel static thrust of engine  [N]

        Outputs:
            WENG: dry engine weight                                         [kg]

        Properties Used:
            N/A
    """
    EEXP = 1.15
    EINL = 1
    ENOZ = 1
    THRSO = ref_propulsor.sealevel_static_thrust * 1 / Units.lbf
    THRUST = THRSO
    if vehicle.systems.accessories == "short-range" or vehicle.systems.accessories == "commuter":
        WENGB = THRSO / 10.5
    else:
        WENGB = THRSO / 5.5
    WINLB = 0 / Units.lbs
    WNOZB = 0 / Units.lbs
    WENGP = WENGB * (THRUST / THRSO) ** EEXP
    WINL = WINLB * (THRUST / THRSO) ** EINL
    WNOZ = WNOZB * (THRUST / THRSO) ** ENOZ
    WENG = WENGP + WINL + WNOZ
    return WENG * Units.lbs



# RCAIDE/Methods/Aerodynamics/Airfoil_Panel_Method/heads_method.py
# 
# 
# Created:  Mar 2024, Niranjan Nanjappa

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports    
from RCAIDE.Framework.Core import Data 

# package imports  
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
# heads_method
# ---------------------------------------------------------------------------------------------------------------------- 
def heads_method(npanel,ncases,ncpts,NU,DEL_0,THETA_0,DELTA_STAR_0,CF_0,ShapeFactor_0,RE_L,TURBULENT_COORD,VE_I,DVE_I,TURBULENT_SURF,tol):
    """ Computes the boundary layer characteristics in turbulent
    flow pressure gradients

    Source:
    Head, M. R., and P. Bandyopadhyay. "New aspects of turbulent boundary-layer structure."
    Journal of fluid mechanics 107 (1981): 297-338.

    Assumptions:
    None  

    Inputs: 
    ncases         - number of cases                                                               [unitless]
    ncpts          - number of control points                                                      [unitless]
    DEL_0          - intital bounday layer thickness                                               [m]
    DELTA_STAR_0   - initial displacement thickness                                                [m]
    CF_0           - initial value of the skin friction coefficient                                [unitless]
    H_0            - initial value of the shape factor                                             [unitless]
    THETA_0        - initial momentum thickness                                                    [m]
    TURBULENT_SURF - normalized length of surface                                                  [unitless]
    RE_L           - Reynolds number                                                               [unitless]
    TURBULENT_COORD- x coordinate on surface of airfoil                                            [unitless] 
    VE_I           - boundary layer velocity at all panels                                         [m/s-m] 
    DVE_I          - derivative of boundary layer velocity at all panels                           [m/s^2] 
    npanel         - number of points on surface                                                   [unitless]
    tol            - boundary layer error correction tolerance                                     [unitless]

    Outputs: 
    RESULTS.
      X_H          - reshaped distance along airfoil surface                    [unitless]
      THETA_H      - momentum thickness                                         [m]
      DELTA_STAR_H - displacement thickness                                     [m] 
      H_H          - shape factor                                               [unitless]
      CF_H         - friction coefficient                                       [unitless]
      RE_THETA_H   - Reynolds number as a function of momentum thickness        [unitless]
      RE_X_H       - Reynolds number as a function of distance                  [unitless]
      DELTA_H      - boundary layer thickness                                   [m]
       
    Properties Used:
    N/A
    """    
    # Initialize vectors 
    X_H          = np.zeros((npanel,ncases,ncpts))
    THETA_H      = np.zeros_like(X_H)
    DELTA_STAR_H = np.zeros_like(X_H)
    H_H          = np.zeros_like(X_H)
    CF_H         = np.zeros_like(X_H) 
    RE_THETA_H   = np.zeros_like(X_H)
    RE_X_H       = np.zeros_like(X_H)
    DELTA_H      = np.zeros_like(X_H)      
    
    for case in range(ncases):
        for cpt in range(ncpts):  
            # length of tubulent surface  
            l = TURBULENT_SURF[case,cpt] 
            if l == 0.0:
                pass
            else: 
                def getcf(ind, nu, H, THETA):
                    ReTheta = Ve_i[ind]*THETA/nu;
                    cf_var = 0.246*(10**(-0.678*H))*(ReTheta**-0.268);
                    return cf_var
                
                def getH(H1_var):
                    if H1_var<3.3:
                        H_var = 3.0
                    elif H1_var < 5.39142:
                        H_var = 0.6778 + 1.153793*(H1_var-3.3)**-0.32637;
                    elif H1_var >= 5.39142:
                        H_var = 1.1 + 0.8598636*(H1_var - 3.3)**-0.777;
                    return H_var 
             
                # define RK4 slope function for Theta
                def dTheta_by_dx(index, X, THETA, VETHETAH1):
                    return 0.5*cf[index] - (THETA/Ve_i[index])*(2+H[index])*(dVe_i[index])
                
                # define RK4 slope function for VeThetaH1
                def dVeThetaH1_by_dx(index, X, THETA, VETHETAH1):
                    return Ve_i[index]*0.0306*(((VETHETAH1/(Ve_i[index]*THETA))-3)**-0.6169)
                
                x_i          = TURBULENT_COORD.data[:,case,cpt][TURBULENT_COORD.mask[:,case,cpt] ==False] 
                Ve_i         = VE_I.data[:,case,cpt][TURBULENT_COORD.mask[:,case,cpt] ==False]
                dVe_i        = DVE_I.data[:,case,cpt][TURBULENT_COORD.mask[:,case,cpt] ==False]
                Re_L         = RE_L[case,cpt]
                n            = len(x_i)
                dx           = np.diff(x_i)
                nu           = NU[case,cpt]
                
                H            = np.zeros(n) 
                H[0]         = ShapeFactor_0[case,cpt]
                Theta        = np.zeros(n)
                Theta[0]     = THETA_0[case,cpt]
                H1           = np.zeros(n) 
                H1[0]        = (DEL_0[case,cpt] - DELTA_STAR_0[case,cpt])/THETA_0[case,cpt]
                if H1[0]<3.3:
                    H1[0] = 3.417285
                
                cf           = np.zeros(n)
                cf[0]        = CF_0[case,cpt] 
                VeThetaH1    = np.zeros(n)
                VeThetaH1[0] = Ve_i[0]*Theta[0]*H1[0]
                
                for i in range(1,n):
                    # initialise the variable values at the current grid point using previous grid points (to define the error functions)
                    H_er = H[i-1];  cf_er = cf[i-1];  H1_er = H1[i-1];  Theta_er = Theta[i-1];
                    # assign previous grid point values of H and Cf to start RK4
                    H[i] = H[i-1]; cf[i] = cf[i-1];
                    
                    #assume some error values
                    erH = 0.2; erH1 = 0.2; erTheta = 0.2; ercf = 0.2;
                    
                    # iterate to get the variables at the grid point
                    while abs(erH)>0.00001 or abs(erH1)>0.00001 or abs(erTheta)>0.00001 or abs(ercf)>0.00001:
                        
                        # get Theta and VeThetaH1
                        Theta[i], VeThetaH1[i] = RK4(i-1, dx, x_i, Theta, VeThetaH1, dTheta_by_dx, dVeThetaH1_by_dx)
                        if np.isnan(VeThetaH1[i]):
                            VeThetaH1[i] = VeThetaH1[i-1]
                       
                        # get H1
                        H1[i] = VeThetaH1[i]/(Ve_i[i]*Theta[i])
                        
                        # get H
                        H[i] = getH(H1[i])
                        
                        # get skin friction
                        cf[i] = getcf(i, nu, H[i], Theta[i])
                        
                        # define errors
                        erH = (H[i]-H_er)/H[i];
                        erH1 = (H1[i]-H1_er)/H1[i];
                        erTheta = (Theta[i]-Theta_er)/Theta[i];
                        ercf = (cf[i]-cf_er)/cf[i];
                        
                        # assign current iteration variable values to the Var_er
                        H_er = H[i]
                        H1_er = H1[i]
                        Theta_er = Theta[i]
                        cf_er = cf[i]
                
                
                delta_star   = H*Theta
                Re_theta     = Ve_i*Theta/nu
                Re_x         = (Ve_i*x_i)/nu
                delta        = (Theta*H1) + delta_star
                
                indices = np.where(TURBULENT_COORD.mask[:,case,cpt] == False)
                np.put(X_H[:,case,cpt],indices,x_i )
                np.put(THETA_H[:,case,cpt],indices,Theta)
                np.put(DELTA_STAR_H[:,case,cpt],indices,delta_star)
                np.put(H_H[:,case,cpt],indices,H)
                np.put(CF_H[:,case,cpt],indices,cf)
                np.put(RE_THETA_H[:,case,cpt],indices,Re_theta)
                np.put(RE_X_H[:,case,cpt],indices,Re_x)
                np.put(DELTA_H[:,case,cpt],indices,delta)

    RESULTS = Data(
            X_H          = X_H,      
            THETA_H      = THETA_H,   
            DELTA_STAR_H = DELTA_STAR_H,
            H_H          = H_H,       
            CF_H         = CF_H,   
            RE_THETA_H   = RE_THETA_H,   
            RE_X_H       = RE_X_H,    
            DELTA_H      = DELTA_H,   
        )    

    return  RESULTS


def RK4(ind, dx, x, Theta_var, VeThetaH1_var, Theta_slope, VeThetaH1_slope):
    k1 = Theta_slope(ind,  x[ind],  Theta_var[ind],  VeThetaH1_var[ind])
    l1 = VeThetaH1_slope(ind,  x[ind],  Theta_var[ind],  VeThetaH1_var[ind])
    
    k2 = Theta_slope(ind,  x[ind] + (dx[ind]/2),  Theta_var[ind] + (k1*dx[ind]/2),  VeThetaH1_var[ind] + (l1*dx[ind]/2))
    l2 = VeThetaH1_slope(ind,  x[ind] + (dx[ind]/2),  Theta_var[ind] + (k1*dx[ind]/2),  VeThetaH1_var[ind] + (l1*dx[ind]/2))
    
    k3 = Theta_slope(ind,  x[ind] + (dx[ind]/2),  Theta_var[ind] + (k2*dx[ind]/2),  VeThetaH1_var[ind] + (l2*dx[ind]/2))
    l3 = VeThetaH1_slope(ind,  x[ind] + (dx[ind]/2),  Theta_var[ind] + (k2*dx[ind]/2),  VeThetaH1_var[ind] + (l2*dx[ind]/2))
    
    k4 = Theta_slope(ind,  x[ind] + dx[ind],  Theta_var[ind] + (k3*dx[ind]),  VeThetaH1_var[ind] + (l2*dx[ind]))
    l4 = VeThetaH1_slope(ind,  x[ind] + dx[ind],  Theta_var[ind] + (k3*dx[ind]),  VeThetaH1_var[ind] + (l2*dx[ind]))
    
    Theta_new = Theta_var[ind] + ((dx[ind]/6)*(k1 + 2*k2 + 2*k3 + k4))
    VeThetaH1_new = VeThetaH1_var[ind] + ((dx[ind]/6)*(l1 + 2*l2 + 2*l3 + l4))
    return Theta_new, VeThetaH1_new 


# RCAIDE/Library/Missions/Segments/Descent/ Constant_Speed_Constant_Angle_noise.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------  
#  IMPORT 
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports  
from RCAIDE.Framework.Core import Units 

# pacakge imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Expand State
# ----------------------------------------------------------------------------------------------------------------------   
def expand_state(segment):
    """Makes all vectors in the state the same size.

    Assumptions:
    A 4 km threshold, this discretizes the mission to take measurements at the right place for certification maneuvers.

    Source:
    N/A

    Inputs:
    state.numerics.number_of_control_points  [Unitless]
    segment.descent_angle                 [Radians]
    segment.air_speed                     [meters/second]
 
    Outputs:
    state.numerics.number_of_control_points

    Properties Used:
    N/A
    """         
    dt            = 0.5  # time step in seconds for noise calculation - Certification requirement    
    air_speed     = segment.air_speed  
    s0            = 4000. # Defining the initial position of the measureament will start at 4 km from the threshold
    v_x           = air_speed * np.cos(segment.descent_angle) 
    
    #number of time steps (space discretization)  
    total_time    = s0/v_x    
    n_points      = np.ceil(total_time/dt +1)       
    
    segment.state.numerics.number_of_control_points = n_points
    segment.state.expand_rows(int(n_points),override=True)   
    
    return

# ----------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------

def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Constant speed, constant descent angle. However, this follows a 2000 meter segment. This is a certification maneuver standard. The last point for the noise measurement is 50 feet.

    Source:
    N/A

    Inputs:
    segment.descent_angle                       [radians]
    segment.altitude_start                      [meters]
    segment.altitude_end                        [meters]
    segment.air_speed                           [meters/second]
    state.numerics.dimensionless.control_points [array]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]
    conditions.frames.inertial.time             [seconds]

    Properties Used:
    N/A
    """     
    
    
    # unpack
    descent_angle= segment.descent_angle
    air_speed    = segment.air_speed  
    beta         = segment.sideslip_angle 
    t_nondim     = segment.state.numerics.dimensionless.control_points
    conditions   = segment.state.conditions  
    
    altf = 50. * Units.feet #(50ft last point for the noise measureament)
    
    #Linear equation: y-y0=m(x-x0)
    m_xx0 = 2000 * np.tan(descent_angle)
    y0    =  m_xx0 + altf  #(Altitude at the microphone X position)
    
    alt0 = y0 + m_xx0 #(Initial altitude of the aircraft)

    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0 
    
    # check for initial velocity vector
    if air_speed is None:
        if not segment.state.initials: raise AttributeError('initial airspeed not set')
        air_speed  =  np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1,:])     
        
    # process velocity vector
    v_mag = air_speed
    v_x   = np.cos(beta) * v_mag * np.cos(-descent_angle)
    v_y   = np.sin(beta) * v_mag * np.cos(-descent_angle)
    v_z   = -v_mag * np.sin(-descent_angle)
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0] = v_x
    conditions.frames.inertial.velocity_vector[:,1] = v_y
    conditions.frames.inertial.velocity_vector[:,2] = v_z
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down
    conditions.freestream.altitude[:,0]             =  alt[:,0] # positive altitude in this context    
    



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Common/compute_main_wing_weight.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Framework.Core import  Units 
import numpy as  np 
# ---------------------------------------------------------------------------------------------------------------------- 
# Main Wing 
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_main_wing_weight(vehicle, wing, rho, sigma):
    """ Calculate the wing weight of the aircraft based on the fully-stressed
    bending weight of the wing box

    Assumptions:
        calculated total wing weight based on a bending index and actual data
        from 15 transport aircraft

    Source:
        http://aerodesign.stanford.edu/aircraftdesign/AircraftDesign.html
        search for: Derivation of the Wing Weight Index

    Inputs:
        vehicle - data dictionary with vehicle properties                   [dimensionless]
            -.mass_properties.max_takeoff: MTOW                             [kilograms]
            -.mass_properties.max_zero_fuel: zero fuel weight aircraft      [kilograms]
            -.flight_envelope.ultimate_load: ultimate load factor
        wing    - data dictionary with specific wing properties             [dimensionless]
            -.areas.reference: wing reference surface area                  [m^2]
            -.sweeps.quarter_chord: quarter chord sweep angle               [deg]
            -.spans.projected: wing span                                    [m]
            -.thickness_to_chord: thickness to chord of wing
            -.taper: taper ratio of wing
            -.chords.root: root chord                                       [m]

    Outputs:
        weight - weight of the wing                  [kilograms]

    Properties Used:
        N/A
    """

    # unpack inputs
    span        = wing.spans.projected
    taper       = wing.taper
    sweep       = wing.sweeps.quarter_chord
    area        = wing.areas.reference
    t_c_w       = wing.thickness_to_chord
    RC          = wing.chords.root
    frac        = wing.areas.reference / vehicle.reference_area
    rho_sigma   = rho * 9.81 / sigma
    Nult        = vehicle.flight_envelope.ultimate_load
    TOW         = vehicle.mass_properties.max_takeoff
    W_zf        = vehicle.mass_properties.max_zero_fuel

    # Start the calculations
    l_tot   = Nult * np.sqrt(TOW * W_zf) * 9.81
    L0      = frac * 2 * l_tot / (span * np.pi)
    

    segment_keys   = list(wing.segments.keys())
    num_segments   = len(segment_keys)     
    if num_segments > 0: 
        # Prime some numbers
        run_sum = 0
        b       = span
        for i_seg in range(1,num_segments): 
            # Unpack segment level info
            Y1 = wing.segments[segment_keys[i_seg-1]].percent_span_location
            Y2 = wing.segments[segment_keys[i_seg]].percent_span_location

            if wing.segments[segment_keys[i_seg-1]].root_chord_percent == wing.segments[segment_keys[i_seg]].root_chord_percent and \
                    wing.segments[segment_keys[i_seg-1]].thickness_to_chord == wing.segments[segment_keys[i_seg]].thickness_to_chord:
                C   = wing.segments[segment_keys[i_seg]].root_chord_percent * RC
                G   = wing.segments[segment_keys[i_seg]].thickness_to_chord
                SW  = wing.segments[segment_keys[i_seg-1]].sweeps.quarter_chord

                WB = (1 / (G * C * np.cos(SW) ** 2)) * 1 / 3 * (
                        1 / 8 * (-Y1 * (5 - 2 * Y1 ** 2) * np.sqrt(1 - Y1 ** 2) -
                                 3 * np.arcsin(Y1)) + 1 / 8 * (
                                Y2 * (5 - 2 * Y2 ** 2) * np.sqrt(1 - Y2 ** 2) + 3 * np.arcsin(Y2)))

            else:
                # A is the root thickness
                A = RC * wing.segments[segment_keys[i_seg-1]].root_chord_percent * wing.segments[segment_keys[i_seg-1]].thickness_to_chord
                # B is the slope of the thickness
                B = (A - RC * wing.segments[segment_keys[i_seg]].root_chord_percent * wing.segments[segment_keys[i_seg]].thickness_to_chord) / (
                        wing.segments[segment_keys[i_seg]].percent_span_location - wing.segments[segment_keys[i_seg-1]].percent_span_location)
                # C is the offset
                C = wing.segments[segment_keys[i_seg-1]].percent_span_location
                SW = wing.segments[segment_keys[i_seg-1]].sweeps.quarter_chord

                WB1 = big_integral(Y1, A, B, C)
                WB2 = big_integral(Y2, A, B, C)

                WB = (WB2 - WB1) / (np.cos(SW) ** 2)

            run_sum += np.real(WB)

        weight_factor = rho_sigma * (b ** 2) * L0 * run_sum / 2

        weight = 4.22 * area / Units.feet ** 2 + (weight_factor / Units.lb)

    else:

        area    = wing.areas.reference / Units.ft ** 2
        span    = wing.spans.projected / Units.ft
        mtow    = TOW / Units.lb  # Convert kg to lbs
        zfw     = W_zf / Units.lb  # Convert kg to lbs

        # Calculate weight of wing for traditional aircraft wing
        weight  = 4.22 * area + 1.642 * 10. ** -6. * Nult * (span) ** 3. * (mtow * zfw) ** 0.5 \
                 * (1. + 2. * taper) / (t_c_w * (np.cos(sweep)) ** 2. * area * (1. + taper))


    weight = weight * Units.lb  # Convert lb to kg

    return weight


def big_integral(x, A, B, C):
    """ Integrate the wing bending moment over a section

    Assumptions:
        Linearly tapering thickness

    Source:
        Botero 2019

    Inputs:
        x - span wise station      [dimensionless]
        A - origin thickness       [meters]
        B - taper ratio of section [dimensionless]
        C - origin offset          [dimensionless]

    Outputs:
        result - evaluate one side of an indefinite integral [meters^-1]

    Properties Used:
        N/A
    """

    results = (1/(4*B**3))*(2*A**2*(np.pi)*x+4*A*B*C*(np.pi)*x+ \
            B**2*(-1+2*C**2)*(np.pi)*x- \
            2*(2*A**2+4*A*B*C+B**2*(-1+2*C**2))*np.sqrt(1-x**2)+ \
            2/3*B*np.sqrt(1-x**2)*(3*A*x+B*(-1+3*C*x+x**2))+ \
            2*B*(A+B*C)*np.arcsin(x)- \
            2*(2*A**2+4*A*B*C+B**2*(-1+2*C**2))*x*np.arcsin(x)-( \
            4*(A+B*C)**2*np.sqrt(0j+-A**2-2*A*B*C-B**2*(-1+C**2))* \
            np.log(A+B*C-B*x))/B-( \
            4*(A**3+3*A**2*B*C+B**3*C*(-1+C**2)+A*B**2*(-1+3*C**2))*x*np.log( \
            A+B*C-B*x))/np.sqrt(0j+-A**2-2*A*B*C-B**2*(-1+C**2))+( \
            8*(A+B*C)**2*np.sqrt(0j-A**2-2*A*B*C-B**2*(-1+C**2))* \
            np.log(0j-A-B*C+B*x))/ \
            B+(4*(A**3+3*A**2*B*C+B**3*C*(-1+C**2)+ \
            A*B**2*(-1+3*C**2))*x*np.log(0j-B+A*x+B*C*x- \
            np.sqrt(0j+-A**2+B**2-2*A*B*C-B**2*C**2)*np.sqrt( \
            1-x**2)))/(np.sqrt(0j-A**2-2*A*B*C-B**2*(-1+C**2)))-(1/B)* \
            4*(A+B*C)**2*np.sqrt(0j-A**2-2*A*B*C-B**2*(-1+C**2))* \
            np.log(-B+A*x+B*C*x+ \
            np.sqrt(0j-A**2+B**2-2*A*B*C-B**2*C**2)*np.sqrt(1-x**2))+(1/B)* 4*(A+B*C)*np.sqrt(0j-(A**2+2*A*B*C+B**2*(-1+C**2))**2)*np.log(A**2*x+2*A*B*C*x+B**2*(-1+C**2)*x+ \
            np.sqrt(0j-(A**2+2*A*B*C+B**2*(-1+C**2))**2)*np.sqrt(1-x**2)))

    return results



# RCAIDE/Library/Missions/Common/Unpack_Unknowns/orientation.py
# 
# 
# Created:  Jul 2023, M. Clarke
# ----------------------------------------------------------------------------------------------------------------------
#  Unpack Unknowns
# ----------------------------------------------------------------------------------------------------------------------
def orientation(segment): 
        
    ctrls    = segment.assigned_control_variables 

    # Body Angle Control 
    if segment.trim_lift_coefficient !=  None:
        segment.state.conditions.aerodynamics.coefficients.lift.total  = segment.trim_lift_coefficient * segment.state.ones_row(1)
    else: 
        if ctrls.body_angle.active: 
            segment.state.conditions.frames.body.inertial_rotations[:,1] = segment.state.unknowns.body_angle[:,0]  
        else: 
            segment.state.conditions.frames.body.inertial_rotations[:,1] = segment.angle_of_attack            



    if ctrls.bank_angle.active: 
        segment.state.conditions.frames.body.inertial_rotations[:,0] = segment.state.unknowns.bank_angle[:,0]
    else:
        segment.state.conditions.frames.body.inertial_rotations[:,0] = segment.bank_angle
        
    segment.state.conditions.frames.body.inertial_rotations[:,2] =  segment.state.conditions.frames.planet.true_heading[:,0] 
    
    # Velocity Control
    if ctrls.velocity.active:
        segment.state.conditions.frames.inertial.velocity_vector[:,0] = segment.state.unknowns.velocity[:,0]
        
    # Altitude Control
    if ctrls.altitude.active:
        segment.state.conditions.frames.inertial.position_vector[:,2] = -segment.state.unknowns.altitude[:,0]
        
    return 
            
            
            


# RCAIDE/Methods/Aerodynamics/Common/Lift/compute_RHS_matrix.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
 
from RCAIDE.Framework.Core import Data 

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Compute RHS matrix 
# ----------------------------------------------------------------------------------------------------------------------    
def compute_RHS_matrix(delta,phi,conditions,settings,geometry,propeller_wake_model):

    """ This computes the right hand side matrix for the VLM. In this
    function, induced velocites from propeller wake are also included
    when relevent and where specified

    Source:
    1. Low-Speed Aerodynamics, Second Edition by Joseph Katz, Allen Plotkin
    Pgs. 331-338

    2. VORLAX Source Code

    Inputs:
    geometry
        networks                                 [Unitless]
        vehicle vortex distribution              [Unitless]

    conditions.aerodynamics.angles.alpha      [radians]
    conditions.aerodynamics.angles.beta     [radians]
    conditions.freestream.velocity               [m/s]
    conditions.static_stability.pitch_rate      [radians/s]
    conditions.static_stability.roll_rate       [radians/s]
    conditions.static_stability.yaw_rate        [radians/s]

    sur_flag    - use_surrogate flag             [Unitless]
    slipstream  - propeller_wake_model flag      [Unitless]
    delta, phi  - flow tangency angles           [radians]

    Outputs:
    rhs.
        RHS
        ONSET
        Vx_ind_total
        Vz_ind_total
        V_distribution
        dt
        YGIRO
        ZGIRO
        VX
        SCNTL
        CCNTL
        COD
        SID

    Properties Used:
    N/A
    """

    # unpack
    VD               = geometry.vortex_distribution

    aoa              = conditions.aerodynamics.angles.alpha
    aoa_distribution = np.repeat(aoa, VD.n_cp, axis = 1)
    PSI              = conditions.aerodynamics.angles.beta
    PSI_distribution = np.repeat(PSI, VD.n_cp, axis = 1)
    V_inf            = conditions.freestream.velocity
    V_distribution   = np.repeat(V_inf , VD.n_cp, axis = 1)
    num_ctrl_pts     = len(aoa)  
    num_eval_pts     = len(VD.XC)

    rot_V_wake_ind   = np.zeros((num_ctrl_pts, VD.n_cp,3))
    Vx_ind_total     = np.zeros_like(V_distribution)
    Vy_ind_total     = np.zeros_like(V_distribution)
    Vz_ind_total     = np.zeros_like(V_distribution) 
    dt               = 0
    for network in geometry.networks:
        if propeller_wake_model:
            rot_V_wake_ind = np.zeros((num_ctrl_pts,num_eval_pts,3))
            for propulsor in network.propulsors: 
                if 'rotor' in  propulsor:
                    rotor =  propulsor.rotor
                elif 'propeller' in  propulsor :
                    rotor =  propulsor.propeller
                rotor_conditions =  conditions.energy[propulsor.tag][rotor.tag]
                rot_V_wake_ind += rotor.Wake.evaluate_slipstream(rotor,rotor_conditions,geometry,num_ctrl_pts) 
                    
            # update the total induced velocity distribution
            Vx_ind_total = Vx_ind_total  + rot_V_wake_ind[:,:,0]
            Vy_ind_total = Vy_ind_total  + rot_V_wake_ind[:,:,1]
            Vz_ind_total = Vz_ind_total  + rot_V_wake_ind[:,:,2]

            rhs = build_RHS(VD, conditions, settings, aoa_distribution, delta, phi, PSI_distribution,
                            Vx_ind_total, Vy_ind_total, Vz_ind_total, V_distribution, dt)           
            
            return  rhs

    rhs = build_RHS(VD, conditions, settings, aoa_distribution, delta, phi, PSI_distribution,
                    Vx_ind_total, Vy_ind_total, Vz_ind_total, V_distribution, dt)
    
    return rhs 

def build_RHS(VD, conditions, settings, aoa_distribution, delta, phi, PSI_distribution,
              Vx_ind_total, Vy_ind_total, Vz_ind_total, V_distribution, dt):
    """ This uses freestream conditions, induced wake velocities, and rotation
    rates (pitch, roll, yaw) to find the boundary conditions (RHS) needed to
    compute gamma. The function defaults to a classical boundary condition
    equation, RHS = V dot N, where V is the unit velocity vector only the panel
    and N is the panel unit normal. However, the user may also define
    settings.use_VORLAX_matrix_calculation to use the boundary condition equation
    from VORLAX, the code on which this VLM is based. This is useful for future
    developers who need to compare numbers 1:1 with VORLAX.

    Note if using VORLAX boundary conditions:
    VORLAX does not take induced wake velocity into account. Additionally,
    VORLAX uses the camber, twist, and dihedral values of a strip's leading
    edge panel for every panel in that strip when calculating panel normals.

    Source:
    1. Low-Speed Aerodynamics, Second Edition by Joseph Katz, Allen Plotkin Pgs. 331-338

    2. VORLAX Source Code

    Inputs:
    settings.use_VORLAX_matrix_calculation  - RHS equation switch               [boolean]
    conditions.static_stability.pitch_rate -                                   [radians/s]
    conditions.static_stability.roll_rate  -                                   [radians/s]
    conditions.static_stability.yaw_rate   -                                   [radians/s]
    aoa_distribution                        - angle of attack                   [radians]
    PSI_distribution                        - sideslip angle                    [radians]
    V[]_ind_total                           - component induced wake velocities [m/s]
    V_distribution                          - freestream velocity magnitude     [m/s]
    delta, phi                              - flow tangency angles              [radians]

    Outputs:
    rhs  -  a Data object used to hold several values including RHS

    Properties Used:
    N/A
    """
    LE_ind      = VD.leading_edge_indices
    RNMAX       = VD.panels_per_strip


    # VORLAX frame RHS calculation---------------------------------------------------------
    #VORLAX subroutine = BOUNDY
    #unpack conditions
    ALFA   = aoa_distribution
    PSIRAD = PSI_distribution
    PITCHQ = conditions.static_stability.pitch_rate
    ROLLQ  = conditions.static_stability.roll_rate
    YAWQ   = conditions.static_stability.yaw_rate
    VINF   = conditions.freestream.velocity

    SINALF = np.sin(ALFA)
    COSIN  = np.cos(ALFA) * np.sin(PSIRAD)
    COSCOS = np.cos(ALFA) * np.cos(PSIRAD)
    PITCH  = PITCHQ / VINF
    ROLL   = ROLLQ  / VINF
    YAW    = YAWQ   / VINF

    #unpack/rename variables from VD
    X      = VD.XCH
    YY     = VD.YCH
    ZZ     = VD.ZCH
    XBAR   = VD.XBAR
    ZBAR   = VD.ZBAR

    CHORD  = VD.chord_lengths[0,:]
    DELTAX = 0.5/RNMAX

    # LOCATE VORTEX LATTICE CONTROL POINT WITH RESPECT TO THE
    # ROTATION CENTER (XBAR, 0, ZBAR). THE RELATIVE COORDINATES
    # ARE XGIRO, YGIRO, AND ZGIRO.
    XGIRO = X + CHORD*DELTAX - np.repeat(XBAR, RNMAX[LE_ind])
    YGIRO = YY
    ZGIRO = ZZ - np.repeat(ZBAR, RNMAX[LE_ind])

    # VX, VY, VZ ARE THE FLOW ONSET VELOCITY COMPONENTS AT THE LEADING
    # EDGE (STRIP MIDPOINT). VX, VY, VZ AND THE ROTATION RATES ARE
    # REFERENCED TO THE FREE STREAM VELOCITY.
    VX = (COSCOS - PITCH*ZGIRO + YAW  *YGIRO)
    VY = (COSIN  - YAW  *XGIRO + ROLL *ZGIRO)
    VZ = (SINALF - ROLL *YGIRO + PITCH*XGIRO)
    
    #COMPUTE DIRECTION COSINES.
    SCNTL  = VD.SLOPE/np.sqrt(1. + VD.SLOPE **2)
    CCNTL  = 1. / np.sqrt(1.0 + SCNTL**2)
    phi_LE = np.repeat(phi[:,LE_ind]  , RNMAX[LE_ind], axis=1)
    COD    = np.cos(phi_LE)
    SID    = np.sin(phi_LE)

    # COMPUTE ONSET FLOW COMPONENT ALONG THE OUTWARD NORMAL TO
    # THE SURFACE AT THE CONTROL POINT, ALOC.
    ALOC  = VX *SCNTL + VY *CCNTL *SID - VZ *CCNTL *COD

    # COMPUTE VELOCITY COMPONENT ALONG X-AXIS INDUCED BY THE RIGID
    # BODY ROTATION, ONSET.
    ONSET = - PITCH *ZGIRO + YAW *YGIRO

    # Body-Frame RHS calculation----------------------------------------------------------
    # Add wake and rotation effects to the freestream
    Vx_rotation       = -PITCHQ*ZGIRO + YAWQ  *YGIRO
    Vy_rotation       = -YAWQ  *XGIRO + ROLLQ *ZGIRO
    Vz_rotation       = -ROLLQ *YGIRO + PITCHQ*XGIRO

    Vx                = V_distribution*np.cos(aoa_distribution)*np.cos(PSI_distribution) + Vx_rotation + Vx_ind_total
    Vy                = V_distribution*np.cos(aoa_distribution)*np.sin(PSI_distribution) + Vy_rotation + Vy_ind_total
    Vz                = V_distribution*np.sin(aoa_distribution)                          + Vz_rotation + Vz_ind_total    
    
    aoa_distribution  = np.arctan(Vz/ np.sqrt(Vx**2 + Vy**2) )
    PSI_distribution  = np.arctan(Vy / Vx)

    # compute RHS: dot(v, panel_normals)
    V_unit_vector    = (np.array([Vx,Vy,Vz])/V_distribution).T
    panel_normals    = VD.normals[:,np.newaxis,:]
    RHS_from_normals = np.sum(V_unit_vector*panel_normals, axis=2).T    

    #pack values--------------------------------------------------------------------------
    use_VORLAX_RHS = settings.use_VORLAX_matrix_calculation

    rhs = Data()
    rhs.RHS            = RHS_from_normals if not use_VORLAX_RHS else ALOC
    rhs.ONSET          = ONSET
    rhs.Vx_ind_total   = Vx_ind_total
    rhs.Vz_ind_total   = Vz_ind_total
    rhs.V_distribution = V_distribution
    rhs.dt             = dt

    #these values will be used later to calculate EFFINC
    rhs.YGIRO  = YGIRO
    rhs.ZGIRO  = ZGIRO
    rhs.VX     = VX
    rhs.SCNTL  = SCNTL
    rhs.CCNTL  = CCNTL
    rhs.COD    = COD
    rhs.SID    = SID

    return rhs



# RCAIDE/Framework/Analyses/Mission/Segments/Cruise/Constant_Throttle_Constant_Altitude.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports  
from RCAIDE.Framework.Mission.Segments.Evaluate  import Evaluate 
from RCAIDE.Framework.Core                       import Units   
from RCAIDE.Library.Mission                      import Common,Segments
from RCAIDE.Framework.Analyses                   import Process  

# ----------------------------------------------------------------------------------------------------------------------
#  Constant_Throttle_Constant_Altitude
# ----------------------------------------------------------------------------------------------------------------------  

class Constant_Throttle_Constant_Altitude(Evaluate):
    """ Vehicle flies at a set throttle setting. Allows a vehicle to do a level acceleration.
    
        Assumptions:
        None
        
        Source:
        None
    """           
    
    
    # ------------------------------------------------------------------
    #   Data Defaults
    # ------------------------------------------------------------------  

    def __defaults__(self):
        """ This sets the default solver flow. Anything in here can be modified after initializing a segment.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """           
        
        # -------------------------------------------------------------------------------------------------------------- 
        # User Inputs
        # -------------------------------------------------------------------------------------------------------------- 
        self.throttle          = None
        self.altitude          = None
        self.air_speed_start   = None
        self.air_speed_end     = 0.0 
        self.true_course       = 0.0 * Units.degrees  

        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission Specific Unknowns and Residuals 
        # --------------------------------------------------------------------------------------------------------------  
        self.state.residuals.final_velocity_error = 0.0
     
        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission specific processes 
        # --------------------------------------------------------------------------------------------------------------    
        initialize                         = self.process.initialize  
        initialize.conditions              = Segments.Cruise.Constant_Throttle_Constant_Altitude.initialize_conditions      
        iterate                            = self.process.iterate             

        # Update Conditions
        iterate.conditions = Process()
        iterate.conditions.differentials         = Common.Update.differentials_time 
        iterate.conditions.velocity              = Segments.Cruise.Constant_Throttle_Constant_Altitude.integrate_velocity 
        iterate.conditions.orientations          = Common.Update.orientations    
        iterate.conditions.acceleration          = Common.Update.acceleration 
        iterate.conditions.angular_acceleration  = Common.Update.angular_acceleration  
        iterate.conditions.altitude              = Common.Update.altitude
        iterate.conditions.atmosphere            = Common.Update.atmosphere
        iterate.conditions.gravity               = Common.Update.gravity
        iterate.conditions.freestream            = Common.Update.freestream
        iterate.conditions.energy                = Common.Update.thrust
        iterate.conditions.aerodynamics          = Common.Update.aerodynamics
        iterate.conditions.stability             = Common.Update.stability
        iterate.conditions.weights               = Common.Update.weights
        iterate.conditions.forces                = Common.Update.forces
        iterate.conditions.moments               = Common.Update.moments
        iterate.conditions.planet_position       = Common.Update.planet_position
        iterate.residuals.flight_dynamics        = Common.Residuals.flight_dynamics
        iterate.residuals.velocity               = Segments.Cruise.Constant_Throttle_Constant_Altitude.solve_velocity
        iterate.unknowns.mission                 = Common.Unpack_Unknowns.orientation  
        iterate.unknowns.acceleration            = Segments.Cruise.Constant_Throttle_Constant_Altitude.unpack_unknowns  

        return


# RCAIDE/Library/Methods/Weights/Buildups/Common/compute_boom_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Library.Attributes.Materials import Bidirectional_Carbon_Fiber, Carbon_Fiber_Honeycomb, Paint, Unidirectional_Carbon_Fiber, Aluminum, Epoxy, Aluminum_Rib

# package imports 
import numpy as np
import copy as cp

# ----------------------------------------------------------------------------------------------------------------------
# Compute wiring weight
# ----------------------------------------------------------------------------------------------------------------------  
def compute_wing_weight(wing,
         config,
         max_thrust,
         num_analysis_points = 10,
         safety_factor = 1.5,
         max_g_load = 3.8,
         moment_to_lift_ratio = 0.02,
         lift_to_drag_ratio = 7,
         forward_web_locations = [0.25, 0.35],
         rear_web_locations = [0.65, 0.75],
         shear_center_location = 0.25,
         margin_factor = 1.2):
    
    """ Calculates the structural mass of a wing for an eVTOL vehicle based on
        assumption of NACA airfoil wing, an assumed L/D, cm/cl, and structural
        geometry. 
        
        Assumptions:
        If no materials are assigned to vehicle model, appropriate assumptions
        are made based on RCAIDE's Solids Attributes library. 
        
        Shear moment is calculated  by adding thrust from each motor to each
        analysis point that's closer to the wing root than the motor. Accomplished
        by indexing Vt according to a boolean mask of the design points that area
        less than or aligned with the motor location under consideration in an
        iterative loop
     
        Sources:
        Project Vahana Conceptual Trade Study

        Inputs:

            wing                          RCAIDE Wing Data Structure           [None]
                 winglet_fraction         winglet fraction                    [Unitless]
                 motor_spanwise_locations spanwise fraction location of motor [Unitless]
            config                        RCAIDE Config Data Structure         [None]
            maxThrust                     Maximum Thrust                      [N]
            numAnalysisPoints             Analysis Points for Sizing          [Unitless]
            safety_factor                 Design Safety Factor                [Unitless]
            max_g_load                    Maximum Accelerative Load           [Unitless]
            moment_to_lift_ratio          Coeff. of Moment to Coeff. of Lift  [Unitless]
            lift_to_drag_ratio            Coeff. of Lift to Coeff. of Drag    [Unitless]
            forward_web_locations         Location of Forward Spar Webbing    [m]
            rear_web_locations            Location of Rear Spar Webbing       [m]
            shear_center                  Location of Shear Center            [m]
            margin_factor                 Allowable Extra Mass Fraction       [Unitless]

        Outputs: 
            weight:                       Wing Mass                           [kg]
    """

    #-------------------------------------------------------------------------------
    # Unpack Inputs
    #-------------------------------------------------------------------------------

    MTOW             = config.mass_properties.max_takeoff
    wingspan         = wing.spans.projected
    chord            = wing.chords.mean_aerodynamic
    thicknessToChord = wing.thickness_to_chord
    wingArea         = wing.areas.reference

    try:
        wingletFraction = wing.winglet_fraction
    except AttributeError:
        wingletFraction = 0.0

    totalWingArea   = 0
    for w in config.wings:
        totalWingArea += w.areas.reference
    liftFraction    = wingArea/totalWingArea

    motor_locs    = []
    for network in config.networks:
        for propulsor in network.propulsors:
            if propulsor.wing_mounted: 
                motor = propulsor.motor                   
                if motor.origin[0][1] >= 0: 
                    motor_locs.append(motor.origin[0][1]) 

    motor_spanwise_locations = np.array(motor_locs)
    N       = num_analysis_points                   # Number of spanwise points
    SF      = safety_factor                         # Safety Factor
    G_max   = max_g_load                            # Maximum G's experienced during climb
    cmocl   = moment_to_lift_ratio                  # Ratio of cm to cl
    LoD     = lift_to_drag_ratio                    # L/D
    fwdWeb  = cp.deepcopy(forward_web_locations)    # Locations of forward spars
    aftWeb  = cp.deepcopy(rear_web_locations)       # Locations of aft spars
    xShear  = shear_center_location                 # Approximate shear center
    grace   = margin_factor                         # Grace factor for estimation 
    nRibs   = len(motor_spanwise_locations) + 2 

    #-------------------------------------------------------------------------------
    # Unpack Material Properties
    #-------------------------------------------------------------------------------

    try:
        torsMat = wing.materials.skin_materials.torsion_carrier
    except AttributeError:
        torsMat = Bidirectional_Carbon_Fiber()
    torsUSS = torsMat.ultimate_shear_strength
    torsDen = torsMat.density
    torsMGT = torsMat.minimum_gage_thickness

    try:
        coreMat = wing.materials.skin_materials.core
    except AttributeError:
        coreMat = Carbon_Fiber_Honeycomb()
    coreDen = coreMat.density
    coreMGT = coreMat.minimum_gage_thickness

    try:
        bendMat = wing.materials.flap_materials.bending_carrier
    except AttributeError:
        bendMat = Unidirectional_Carbon_Fiber()
    bendUTS = bendMat.ultimate_tensile_strength
    bendDen = bendMat.density

    try:
        glueMat = wing.materials.skin_materials.adhesive
    except AttributeError:
        glueMat = Epoxy()
    glueMGT = glueMat.minimum_gage_thickness
    glueDen = glueMat.density

    try:
        coverMat = wing.materials.skin_materials.covering
    except AttributeError:
        coverMat = Paint()
    coverMGT = coverMat.minimum_gage_thickness
    coverDen = coverMat.density

    try:
        ribMat = wing.rib_materials.structural
    except AttributeError:
        ribMat = Aluminum_Rib()
    ribWid = ribMat.minimum_width
    ribMGT = ribMat.minimum_gage_thickness
    ribDen = ribMat.density

    try:
        shearMat = wing.spar_materials.shear_carrier
    except AttributeError:
        shearMat = Bidirectional_Carbon_Fiber()
    shearMGT = shearMat.minimum_gage_thickness
    shearDen = shearMat.density
    shearUSS = shearMat.ultimate_shear_strength


    #-------------------------------------------------------------------------------
    # Airfoil
    #------------------------------------------------------------------------------- 
    NACA        = np.multiply(5 * thicknessToChord, [0.2969, -0.1260, -0.3516, 0.2843, -0.1015])
    coord       = np.unique(fwdWeb+aftWeb+np.linspace(0, 1, N).tolist())[:, np.newaxis]
    coordMAT    = np.concatenate((coord**0.5, coord, coord**2, coord**3, coord**4), axis=1)
    nacaMAT     = coordMAT.dot(NACA)[:, np.newaxis]
    coord       = np.concatenate((coord, nacaMAT), axis=1)
    coord       = np.concatenate((coord[-1:0:-1], coord.dot(np.array([[1., 0.], [0., -1.]]))), axis=0)
    coord[:, 0] = coord[:, 0] - xShear

    #-------------------------------------------------------------------------------
    # Beam Geometry
    #------------------------------------------------------------------------------- 
    x         = np.concatenate((np.linspace(0, 1, N), np.linspace(1, 1+wingletFraction, N)), axis=0)
    x         = x * wingspan/2
    x         = np.sort(np.concatenate((x,motor_spanwise_locations), axis=0))
    dx        = x[1] - x[0]
    N         = np.size(x)
    fwdWeb[:] = [round(locFwd - xShear, 2) for locFwd in fwdWeb]
    aftWeb[:] = [round(locAft - xShear, 2) for locAft in aftWeb]

    #-------------------------------------------------------------------------------
    # Loads
    #------------------------------------------------------------------------------- 
    L  = (1-(x/np.max(x))**2)**0.5           # Assumes Elliptic Lift Distribution
    L0 = 0.5*G_max*MTOW*9.8*liftFraction*SF  # Total Design Lift Force
    L  = L0/np.sum(L[0:-1]*np.diff(x))*L     # Net Lift Distribution 
    T  = L * chord * cmocl                   # Torsion Distribution
    D  = L/LoD                               # Drag Distribution

    #-------------------------------------------------------------------------------
    # Shear/Moments
    #------------------------------------------------------------------------------- 
    Vx = np.append(np.cumsum((D[0:-1]*np.diff(x))[::-1])[::-1], 0)   # Drag Shear
    Vz = np.append(np.cumsum((L[0:-1]*np.diff(x))[::-1])[::-1], 0)   # Lift Shear
    Vt = 0 * Vz                                                      # Initialize Thrust Shear 

    for i in range(np.size(motor_spanwise_locations)):
        Vt[x<=motor_spanwise_locations[i]] = Vt[x<=motor_spanwise_locations[i]] + max_thrust

    Mx = np.append(np.cumsum((Vz[0:-1]*np.diff(x))[::-1])[::-1],0)  # Bending Moment
    My = np.append(np.cumsum(( T[0:-1]*np.diff(x))[::-1])[::-1],0)  # Torsion Moment
    Mz = np.append(np.cumsum((Vx[0:-1]*np.diff(x))[::-1])[::-1],0)  # Drag Moment
    Mt = np.append(np.cumsum((Vt[0:-1]*np.diff(x))[::-1])[::-1],0)  # Thrust Moment
    Mz = np.max((Mz, Mt))                                           # Worst Case of Drag vs. Thrust Moment

    #-------------------------------------------------------------------------------
    # General Structural Properties
    #------------------------------------------------------------------------------- 
    seg = []                      
    
    # Torsion 
    box = coord                        # Box Initally Matches Airfoil
    box = box[box[:, 0] <= aftWeb[1]]  # Inlcude Only Parts Fwd of Aftmost Spar
    box = box[box[:, 0] >= fwdWeb[0]]  # Include Only Parts Aft of Fwdmost Spar
    box = box * chord                  # Scale by Chord Length

    # Use Shoelace Formula to calculate box area 
    torsionArea = 0.5*np.abs(np.dot(box[:, 0], np.roll(box[:, 1], 1)) -
        np.dot(box[:, 1], np.roll(box[:, 0], 1)))

    torsionLength = np.sum(np.sqrt(np.sum(np.diff(box, axis=0)**2, axis=1)))

    # Bending 
    box = coord                                             # Box Initially Matches Airfoil
    box = box[box[:, 0] <= fwdWeb[1]]                       # Include Only Parts Fwd of Aft Fwd Spar
    box = box[box[:, 0] >= fwdWeb[0]]                       # Include Only Parts Aft of Fwdmost Spar
    seg.append(box[box[:, 1] > np.mean(box[:, 1])]*chord)   # Upper Fwd Segment
    seg.append(box[box[:, 1] < np.mean(box[:, 1])]*chord)   # Lower Fwd Segment

    # Drag 
    box = coord                                             # Box Initially Matches Airfoil
    box = box[box[:, 0] <= aftWeb[1]]                       # Include Only Parts Fwd of Aftmost Spar
    box = box[box[:, 0] >= aftWeb[0]]                       # Include Only Parts Aft of Fwd Aft Spar
    seg.append(box[box[:, 1] > np.mean(box[:, 1])]*chord)   # Upper Aft Segment
    seg.append(box[box[:, 1] < np.mean(box[:, 1])]*chord)   # Lower Aft Segment

    # Bending/Drag Inertia 
    flapInertia = 0
    flapLength  = 0
    dragInertia = 0
    dragLength  = 0

    for i in range(0, 4):
        l = np.sqrt(np.sum(np.diff(seg[i], axis=0)**2, axis=1))    # Segment lengths
        c = (seg[i][1::]+seg[i][0:-1])/2                         # Segment centroids

        if i<2:
            flapInertia += np.abs(np.sum(l*c[:,1]**2))   # Bending Inertia per Unit Thickness
            flapLength  += np.sum(l)
        else:
            dragInertia += np.abs(np.sum(l*c[:,0]**2))   # Drag Inertia per Unit Thickness
            dragLength  += np.sum(l)


    # Shear 
    box        = coord                                                                 # Box Initially Matches Airfoil
    box        = box[box[:,0]<=fwdWeb[1]]                                              # Include Only Parts Fwd of Aft Fwd Spar
    z          = np.zeros(2)
    z[0]       = np.interp(fwdWeb[0], box[box[:, 1] > 0,0],box[box[:,1] > 0,1])*chord  # Upper Surf of Box at Fwdmost Spar
    z[1]       = np.interp(fwdWeb[0], box[box[:, 1] < 0,0],box[box[:,1] < 0,1])*chord  # Lower Surf of Box at Fwdmost Spar
    h          = np.abs(z[0] - z[1])                                                   # Height of Box at Fwdmost Spar

    # Skin 
    box        = coord * chord                                                   # Box Initially is Airfoil Scaled by Chord
    skinLength = np.sum(np.sqrt(np.sum(np.diff(box, axis=0)**2, axis=1)))
    A          = 0.5*np.abs(np.dot(box[:,0],np.roll(box[:, 1], 1)) -
                 np.dot(box[:, 1], np.roll(box[:, 0], 1)))                       # Box Area via Shoelace Formula

    #---------------------------------------------------------------------------
    # Structural Calculations
    #---------------------------------------------------------------------------

    # Calculate Skin Weight Based on Torsion 
    tTorsion = My*dx/(2*torsUSS*torsionArea)                # Torsion Skin Thickness
    tTorsion = np.maximum(tTorsion,torsMGT*np.ones(N))      # Gage Constraint
    mTorsion = tTorsion * torsionLength * torsDen           # Torsion Mass
    mCore    = coreMGT*torsionLength*coreDen*np.ones(N)     # Core Mass
    mGlue    = glueMGT*glueDen*torsionLength*np.ones(N)     # Epoxy Mass

    # Calculate Flap Mass Based on Bending 
    tFlap    = Mx*np.max(seg[0][:,1])/(flapInertia*bendUTS)    # Bending Flap Thickness
    mFlap    = tFlap*flapLength*bendDen                        # Bending Flap Mass
    mGlue    += glueMGT*glueDen*flapLength*np.ones(N)          # Updated Epoxy Mass

    # Calculate Drag Flap Mass 
    tDrag    = Mz*np.max(seg[2][:,0])/(dragInertia*bendUTS)    # Drag Flap Thickness
    mDrag    = tDrag*dragLength*bendDen                        # Drag Flap Mass
    mGlue    += glueMGT*glueDen*dragLength*np.ones(N)          # Updated Epoxy Mass

    # Calculate Shear Spar Mass 
    tShear   = 1.5*Vz/(shearUSS*h)                            # Shear Spar Thickness
    tShear   = np.maximum(tShear, shearMGT*np.ones(N))        # Gage constraint
    mShear   = tShear*h*shearDen                              # Shear Spar Mass

    # Paint 
    mPaint   = skinLength*coverMGT*coverDen*np.ones(N)        # Paint Mass

    # Section Mass Total 
    m    = mTorsion + mCore + mFlap + mDrag + mShear + mGlue + mPaint

    # Rib Mass 
    mRib = (A+skinLength*ribWid)*ribMGT*ribDen

    # Total Mass 
    mass = 2*(sum(m[0:-1]*np.diff(x))+nRibs*mRib)*grace

    return mass


# RCAIDE/Library/Missions/Segments/Transition/Constant_Acceleration_Constant_Pitchrate_Constant_Altitude.py
# 
# 
# Created:  Jul 2023, M. Clarke 
  
# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------   
# Package imports 
import numpy as np

def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Constant acceleration and constant altitude

    Source:
    N/A

    Inputs:
    segment.altitude                [meters]
    segment.air_speed_start         [meters/second]
    segment.air_speed_end           [meters/second]
    segment.acceleration            [meters/second^2]
    conditions.frames.inertial.time [seconds]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]
    conditions.frames.inertial.time             [seconds]

    Properties Used:
    N/A
    """      
    
    # unpack
    alt = segment.altitude 
    v0  = segment.air_speed_start
    vf  = segment.air_speed_end  
    beta= segment.sideslip_angle
    ax  = segment.acceleration   
    T0  = segment.pitch_initial
    Tf  = segment.pitch_final     
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
        segment.altitude = alt
        
    # check for initial pitch
    if T0 is None:
        T0  =  segment.state.initials.conditions.frames.body.inertial_rotations[-1,1]
        segment.pitch_initial = T0    
    
    # check for initial velocity vector
    if v0 is None:
        if not segment.state.initials: raise AttributeError('initial airspeed not set')
        v0  =  segment.state.initials.conditions.frames.inertial.velocity_vector[-1,0] # x direction velocity         

    # dimensionalize time
    t_initial = segment.state.conditions.frames.inertial.time[0,0]
    t_final   = (vf-v0)/ax + t_initial
    t_nondim  = segment.state.numerics.dimensionless.control_points
    time      = t_nondim * (t_final-t_initial) + t_initial
    
    # Figure out vx
    v_mag = v0+(time - t_initial)*ax
    v_x   = np.cos(beta)*v_mag
    v_y   = np.sin(beta)*v_mag
    
    # set the body angle
    if Tf > T0:
        body_angle = T0 + time*(Tf-T0)/(t_final-t_initial)
    else:
        body_angle = T0 - time*(T0-Tf)/(t_final-t_initial)
    segment.state.conditions.frames.body.inertial_rotations[:,1] = body_angle[:,0]     
    
    # pack
    segment.state.conditions.freestream.altitude[:,0] = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = v_x[:,0]
    segment.state.conditions.frames.inertial.velocity_vector[:,1] = v_y[:,0]
    segment.state.conditions.frames.inertial.time[:,0] = time[:,0]



# RCAIDE/Library/Plots/Aerodynamics/plot_aircraft_aerodynamic_analysis.py
# 
# 
# Created:  Dec 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style 
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_aircraft_aerodynamics(results,
                            save_figure = False,
                            line_color = 'bo-',
                            line_color2 = 'rs--',
                            save_filename = "Aircraft_Aerodynamic_Analysis",
                            file_type = ".png",
                            width = 11, height = 7):
    

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
    
    #------------------------------------------------------------------------
    # setup figures
    #------------------------------------------------------------------------
    fig = plt.figure(save_filename + " Lift Coefficients")  
    fig.set_size_inches(12,6) 
    axis_1 = fig.add_subplot(1, 2, 1, projection='3d')
    axis_2 = fig.add_subplot(1, 2, 2, projection='3d') 
 
    X, Y = np.meshgrid(results.Mach, results.alpha)
    surf = axis_1.plot_surface(X, Y/Units.degree, results.lift_coefficient   , cmap=cm.jet,linewidth=0, antialiased=False) 
    surf = axis_2.plot_surface(X, Y/Units.degree, results.drag_coefficient   , cmap=cm.jet,linewidth=0, antialiased=False) 

    axis_1.set_title('$C_L$') 
    axis_2.set_title('$C_L$')            
    axis_1.set_ylabel('AoA') 
    axis_2.set_ylabel('AoA')  
    axis_1.set_xlabel('Mach') 
    axis_2.set_xlabel('Mach')   
    
    # set title of plot 
    title_text    = 'Aircraft Aerodynamic Analysis '    
    fig.suptitle(title_text) 
    
    plt.tight_layout()    
    if save_figure:    
        fig.savefig(save_filename + file_type) 
    
    plt.tight_layout()
    return



# RCAIDE/Compoments/Propulsors/Converters/Ducted_Fan.py
# 
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Core              import Data
from RCAIDE.Library.Components          import Component 
from RCAIDE.Library.Methods.Propulsors.Converters.Ducted_Fan.append_ducted_fan_conditions import  append_ducted_fan_conditions
import numpy as np
import scipy as sp
 
# ---------------------------------------------------------------------------------------------------------------------- 
#  Nacalle
# ----------------------------------------------------------------------------------------------------------------------  
class Ducted_Fan(Component):
    """
    A ducted fan propulsion system model that simulates the performance of a shrouded fan.

    Attributes
    ----------
    tag : str
        Identifier for the ducted fan. Default is 'ducted_fan'.
        
    number_of_radial_stations : int
        Number of radial calculation points for blade analysis. Default is 20.
        
    number_of_rotor_blades : int
        Number of blades in the rotor. Default is 12.
        
    tip_radius : float
        Outer radius of the rotor [m]. Default is 1.0.
        
    hub_radius : float
        Inner radius of the rotor at hub [m]. Default is 0.1.
        
    blade_clearance : float
        Clearance between blade tip and duct [m]. Default is 0.01.
        
    length : float
        Length of the ducted fan assembly [m]. Default is 1.0.
        
    fidelity : str
        Level of fidelity for calculations. Default is 'polytropic'.
        
    nacelle : Component
        Nacelle component. Default is None.
        
    fan : Component
        Fan component. Default is Fan().
        
    ram : Component
        Ram air intake component. Default is Ram().
        
    inlet_nozzle : Component
        Inlet nozzle component. Default is Compression_Nozzle().
        
    orientation_euler_angles : list
        Vector of angles defining default orientation [rad]. Default is [0.,0.,0.].
        
    rotor : Data
        Rotor configuration data

        - percent_x_location : float
            Rotor position as fraction of length. Default is 0.4.
            
    stator : Data
        Stator configuration data

        - percent_x_location : float
            Stator position as fraction of length. Default is 0.7.
            
    cruise : Data
        Design cruise conditions

        - design_thrust : float
            Design point thrust. Default is None.

        - design_altitude : float
            Design altitude [m]. Default is None.
            
        - design_angular_velocity : float
            Design rotational speed [rad/s]. Default is None.

        - design_freestream_velocity : float
            Design forward velocity [m/s]. Default is None.

        - design_reference_velocity : float
            Design reference velocity [m/s]. Default is None.

        - design_freestream_mach : float
            Design Mach number. Default is None.

        - design_reference_mach : float
            Design reference Mach number. Default is None.
            
    duct_airfoil : Data
        Duct aerodynamic surface data. Default is empty Data().
        
    hub_airfoil : Data
        Hub aerodynamic surface data. Default is empty Data().

    Notes
    -----
    The Ducted_Fan class models a shrouded fan propulsion system, including:
    * Rotor-stator interaction effects
    * Duct aerodynamic influences
    * Hub effects
    * Multiple coordinate frame transformations
    
    The model supports various fidelity levels and can handle both design and 
    off-design conditions.

    **Definitions**

    'Euler Angles'
        Set of three angles used to describe orientation in 3D space
        
    'Reference Velocity'
        Characteristic velocity used for non-dimensionalization
        
    'Polytropic'
        Thermodynamic process with constant polytropic efficiency

    See Also
    --------
    RCAIDE.Library.Components.Propulsors.Converters.Fan
    RCAIDE.Library.Components.Propulsors.Converters.Ram
    RCAIDE.Library.Components.Propulsors.Converters.Compression_Nozzle
    """
    
    def __defaults__(self):
        """ This sets the default values for the component to function.
        
        Assumptions:
        None
    
        Source:
        N/A
    
        Inputs:
        None
    
        Outputs:
        None
    
        Properties Used:
        None
        """      
        
        self.tag                               = 'ducted_fan'  
        self.number_of_radial_stations         = 20
        self.number_of_rotor_blades            = 12  
        self.tip_radius                        = 1.0
        self.hub_radius                        = 0.1
        self.blade_clearance                   = 0.01
        self.length                            = 1
        self.fidelity                          = 'polytropic'
        self.nacelle                           = None  
        self.fan                               = RCAIDE.Library.Components.Propulsors.Converters.Fan()   
        self.ram                               = RCAIDE.Library.Components.Propulsors.Converters.Ram()  
        self.inlet_nozzle                      = RCAIDE.Library.Components.Propulsors.Converters.Compression_Nozzle()  
        
        self.orientation_euler_angles          = [0.,0.,0.]  # vector of angles defining default orientation of rotor
        self.rotor                             = Data()
        self.stator                            = Data()
        self.rotor.percent_x_location          = 0.4
        self.stator.percent_x_location         = 0.7
        self.cruise                            = Data()
        self.cruise.design_thrust              = None
        self.cruise.design_altitude            = None
        self.cruise.design_angular_velocity    = None
        self.cruise.design_freestream_velocity = None
        self.cruise.design_reference_velocity  = None 
        self.cruise.design_freestream_mach     = None
        self.cruise.design_reference_mach      = None 
        self.duct_airfoil                      = None
        self.hub_airfoil                       = None
      
    
    def append_duct_airfoil(self, airfoil):
        """
        Adds an airfoil to the ducted fan's duct section.

        Parameters
        ----------
        airfoil : Data
            Airfoil data container with aerodynamic properties for the duct section.
            Must be of type Data().

        Returns
        -------
        None

        Notes
        -----
        This method appends airfoil data to the duct_airfoil attribute of the ducted fan.
        The airfoil data is used to model the aerodynamic characteristics of the duct
        section, which influences the overall performance of the ducted fan system.

        Raises
        ------
        Exception
            If input airfoil is not of type Data()
        """

        # Assert database type
        if not isinstance(airfoil,RCAIDE.Library.Components.Airfoils.Airfoil):
            raise Exception('input component must be of type Airfoil') 

        # Store data
        self.duct_airfoil =  airfoil

        return
    

    def append_hub_airfoil(self, airfoil):
        """
        Adds an airfoil to the ducted fan's hub section.

        Parameters
        ----------
        airfoil : Data
            Airfoil data container with aerodynamic properties for the hub section.
            Must be of type Data().

        Returns
        -------
        None

        Notes
        -----
        This method appends airfoil data to the hub_airfoil attribute of the ducted fan.
        The airfoil data is used to model the aerodynamic characteristics of the hub
        section, which affects the flow field and performance of the ducted fan system.

        Raises
        ------
        Exception
            If input airfoil is not of type Data()
        """

        # Assert database type
        if not isinstance(airfoil,RCAIDE.Library.Components.Airfoils.Airfoil):
            raise Exception('input component must be of type Airfoil') 

        # Store data
        self.hub_airfoil =  airfoil

        return 

    def append_operating_conditions(ducted_fan,segment,propulsor): 
        energy_conditions       = segment.state.conditions.energy[propulsor.tag]
        append_ducted_fan_conditions(ducted_fan,segment,energy_conditions)
        return        
          
    def vec_to_vel(self):
        """
        Rotates from the ducted fan's vehicle frame to the ducted fan's velocity frame.

        Parameters
        ----------
        None

        Returns
        -------
        rot_mat : ndarray
            3x3 rotation matrix transforming from vehicle frame to velocity frame.

        Notes
        -----
        This method creates a rotation matrix for transforming coordinates between
        two reference frames of the ducted fan. When the ducted fan is axially
        aligned with the vehicle body:

        Velocity frame:
        * X-axis points out the nose
        * Z-axis points towards the ground
        * Y-axis points out the right wing

        Vehicle frame:
        * X-axis points towards the tail
        * Z-axis points towards the ceiling
        * Y-axis points out the right wing

        **Theory**
        The transformation is accomplished using a rotation of π radians about the Y-axis,
        represented as a rotation vector [0, π, 0].

        **Major Assumptions**
        * The ducted fan's default orientation is aligned with the vehicle body
        * Right-handed coordinate system is used
        """

        rot_mat = sp.spatial.transform.Rotation.from_rotvec([0,np.pi,0]).as_matrix()

        return rot_mat
    

    def body_to_prop_vel(self, commanded_thrust_vector):
        """
        Rotates from the system's body frame to the ducted fan's velocity frame.

        Parameters
        ----------
        commanded_thrust_vector : ndarray
            Vector of commanded thrust angles [rad] for each time step.

        Returns
        -------
        rot_mat : ndarray
            3x3 rotation matrix transforming from body frame to ducted fan velocity frame.
        rots : ndarray
            Array of rotation vectors including commanded thrust angles.

        Notes
        -----
        This method performs a sequence of rotations to transform coordinates from
        the vehicle body frame to the ducted fan's velocity frame. The transformation
        sequence is:
        1. Body to vehicle frame (π rotation about Y-axis)
        2. Vehicle to ducted fan vehicle frame (includes thrust vector rotation)
        3. Ducted fan vehicle to ducted fan velocity frame

        Reference frames:
        Velocity frame:
        * X-axis points out the nose
        * Z-axis points towards the ground
        * Y-axis points out the right wing

        Vehicle frame:
        * X-axis points towards the tail
        * Z-axis points towards the ceiling
        * Y-axis points out the right wing

        **Theory**
        The complete transformation is computed as:
        rot_mat = (body_2_vehicle @ vehicle_2_duct_vec) @ duct_vec_2_duct_vel

        **Major Assumptions**
        * The ducted fan's default orientation is defined by orientation_euler_angles
        * Right-handed coordinate system is used
        * Thrust vector rotation is applied about the Y-axis
        * Matrix multiplication order preserves proper transformation sequence
        """

        # Go from velocity to vehicle frame
        body_2_vehicle = sp.spatial.transform.Rotation.from_rotvec([0,np.pi,0]).as_matrix()

        # Go from vehicle frame to ducted fan vehicle frame: rot 1 including the extra body rotation
        cpts       = len(np.atleast_1d(commanded_thrust_vector))
        rots       = np.array(self.orientation_euler_angles) * 1.
        rots       = np.repeat(rots[None,:], cpts, axis=0) 
        rots[:,1] += commanded_thrust_vector[:,0] 
        
        vehicle_2_duct_vec = sp.spatial.transform.Rotation.from_rotvec(rots).as_matrix()

        # GO from the ducted fan vehicle frame to the ducted fan velocity frame: rot 2
        duct_vec_2_duct_vel = self.vec_to_vel()

        # Do all the matrix multiplies
        rot1    = np.matmul(body_2_vehicle,vehicle_2_duct_vec)
        rot_mat = np.matmul(rot1,duct_vec_2_duct_vel)
 
        return rot_mat , rots


    def duct_vel_to_body(self, commanded_thrust_vector):
        """
        Rotates from the ducted fan's velocity frame to the system's body frame.

        Parameters
        ----------
        commanded_thrust_vector : ndarray
            Vector of commanded thrust angles [rad] for each time step.

        Returns
        -------
        rot_mat : ndarray
            3x3 rotation matrix transforming from ducted fan velocity frame to body frame.
        rots : ndarray
            Array of rotation vectors including commanded thrust angles.

        Notes
        -----
        This method performs the inverse transformation sequence of body_to_prop_vel.
        The transformation sequence is:
        1. Ducted fan velocity to ducted fan vehicle frame
        2. Ducted fan vehicle to vehicle frame (includes thrust vector rotation)
        3. Vehicle to body frame (π rotation about Y-axis)

        Reference frames:
        Velocity frame:
        * X-axis points out the nose
        * Z-axis points towards the ground
        * Y-axis points out the right wing

        Vehicle frame:
        * X-axis points towards the tail
        * Z-axis points towards the ceiling
        * Y-axis points out the right wing

        **Theory**
        The transformation is computed by inverting the rotation matrix from 
        body_to_prop_vel using:
        rot_mat = (body_2_duct_vel)^(-1)

        **Major Assumptions**
        * The ducted fan's default orientation is defined by orientation_euler_angles
        * Right-handed coordinate system is used
        * Thrust vector rotation is applied about the Y-axis
        """

        body2ductvel,rots = self.body_to_duct_vel(commanded_thrust_vector)

        r = sp.spatial.transform.Rotation.from_matrix(body2ductvel)
        r = r.inv()
        rot_mat = r.as_matrix()

        return rot_mat, rots
    
    def vec_to_duct_body(self,commanded_thrust_vector):
        rot_mat, rots =  self.duct_vel_to_body(commanded_thrust_vector) 
        return rot_mat, rots 


# RCAIDE/Library/Missions/Segments/Common/Initialize/inertial_position.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Inertial Position
# ----------------------------------------------------------------------------------------------------------------------
def inertial_position(segment): 
    """ Initializes intertial positon of vehicle
    
        Assumptions:  
            Only used if there is an initial condition
            
        Inputs: 
            state.conditions:           
                numerics.dimensionless.integrate        [i]
                numerics.dimensionless.control_points   [-]
                frames.inertial.aircraft_range          [m]
                inertial.velocity_vector                [m/s]
            
        Outputs: 
            state.conditions.frames.inertial.position_vector  [m]
            state.conditions.frames.inertial.aircraft_range   [m]
           
        Properties Used:
        N/A 
    """      
    if segment.state.initials:
        r_initial = segment.state.initials.conditions.frames.inertial.position_vector
        r_current = segment.state.conditions.frames.inertial.position_vector
        R_initial = segment.state.initials.conditions.frames.inertial.aircraft_range
        R_current = segment.state.conditions.frames.inertial.aircraft_range
        
        if 'altitude' in segment.keys() and segment.altitude is not None:
            r_initial[-1,None,-1] = -segment.altitude
        elif 'altitude_start' in segment.keys() and segment.altitude_start is not None:
            r_initial[-1,None,-1] = -segment.altitude_start
        else:
            assert('Altitude not set')
            
        segment.state.conditions.frames.inertial.position_vector[:,:] = r_current + (r_initial[-1,None,:] - r_current[0,None,:])
        segment.state.conditions.frames.inertial.aircraft_range[:,:]  = R_current + (R_initial[-1,None,:] - R_current[0,None,:])
        
    return 


# RCAIDE/Library/Methods/Aerodynamics/Common/Drag/total_drag.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  Wave Drag 
# ---------------------------------------------------------------------------------------------------------------------- 
def total_drag(state,settings,geometry):
    """ Computes the total drag of an aircraft.

    Assumptions:
        None

    Source:
        None 

    Args:
        settings.
          drag_coefficient_increment                          (float): drag_coefficient_increment [Unitless]
          lift_to_drag_adjustment                             (float): lift_to_drag_adjustment    [Unitless]  
        state.conditions.aerodynamics.coefficients.drag.
          trim_corrected_drag                         (numpy.ndarray): trim corrected drag        [Unitless]
          spoiler_drag                                (numpy.ndarray): spoiler drag               [Unitless]
        geometry                                               (dict): aircraft data structure    [-]


    Returns:
        None 
    """

    # unpack inputs  
    trim_correction_factor  = settings.trim_drag_correction_factor    
    drag                    =  state.conditions.aerodynamics.coefficients.drag

    # various drag components
    parasite_total        = drag.parasite.total            
    induced_total         = drag.induced.total            
    compressibility_total = drag.compressible.total     
    miscellaneous_drag    = drag.miscellaneous.total
    cooling_drag          = drag.cooling.total 

    # untrimmed drag 
    untrimmed_drag  =  parasite_total + induced_total  + compressibility_total + miscellaneous_drag + cooling_drag
    
    # trim correction
    corrected_aircraft_total_trim_drag = trim_correction_factor * untrimmed_drag 
    drag.miscellaneous.trim_correction_factor = trim_correction_factor    

    # Unpack inputs 

    drag_coefficient_increment = settings.drag_coefficient_increment 
    spoiler_drag               = drag.spoiler_drag 

    # Add drag_coefficient_increment
    aircraft_total_drag =  corrected_aircraft_total_trim_drag   + drag_coefficient_increment + spoiler_drag
    drag.drag_coefficient_increment = drag_coefficient_increment
    
    # Add L/D correction
    aircraft_total_drag  = aircraft_total_drag/(1.+settings.lift_to_drag_adjustment) 

    # Store to results 
    drag.total           = aircraft_total_drag
    
    return  


# RCAIDE/Library/Methods/Aerdoynamics/AERODAS/post_stall_coefficients.py
# 
# 
# Created:  Jul 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
# Imports 
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Framework.Core import Data, Units

# python imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
# Post Stall Coefficients
# ----------------------------------------------------------------------------------------------------------------------    
def post_stall_coefficients(state,settings,geometry):
    """Uses the AERODAS method to determine poststall parameters for lift and drag for a single wing

    Assumptions:
    None

    Source:
    NASA TR: "Models of Lift and Drag Coefficients of Stalled and Unstalled Airfoils in
      Wind Turbines and Wind Tunnels" by D. A. Spera

    Inputs:
    settings.section_zero_lift_angle_of_attack      [radians]
    geometry.
      aspect_ratio                                  [Unitless]
      thickness_to_chord                            [Unitless]
      section.angle_attack_max_prestall_lift        [radians]
      pre_stall_maximum_lift_drag_coefficient       [Unitless]
      pre_stall_maximum_drag_coefficient_angle      [Unitless]
    state.conditions.aerodynamics.angle_of_attack   [radians]
      

    Outputs:
    CL2 (coefficient of lift)                       [Unitless]
    CD2 (coefficient of drag)                       [Unitless]
    (packed in state.conditions.aerodynamics.post_stall_coefficients[geometry.tag])

    Properties Used:
    N/A
    """  
    
    # unpack inputs
    wing   = geometry
    A0     = settings.section_zero_lift_angle_of_attack
    AR     = wing.aspect_ratio
    t_c    = wing.thickness_to_chord
    ACL1   = wing.section.angle_attack_max_prestall_lift 
    CD1max = wing.pre_stall_maximum_lift_drag_coefficient
    ACD1   = wing.pre_stall_maximum_drag_coefficient_angle
    alpha  = state.conditions.aerodynamics.angle_of_attack
    
    if wing.vertical == True:
        alpha = 0. * np.ones_like(alpha)    
            
    # Equation 9a and b
    F1        = 1.190*(1.0-(t_c*t_c))
    F2        = 0.65 + 0.35*np.exp(-(9.0/AR)**2.3)
    
    # Equation 10b and c
    G1        = 2.3*np.exp(-(0.65*t_c)**0.9)
    G2        = 0.52 + 0.48*np.exp(-(6.5/AR)**1.1)
    
    # Equation 8a and b
    CL2max    = F1*F2
    CD2max    = G1*G2
    
    # Equation 11d
    RCL2      = 1.632-CL2max
    
    # Equation 11e
    N2        = 1 + CL2max/RCL2
    
    # Equation 11a,b,c
    con1      = np.logical_and(0<alpha,alpha<ACL1)
    con2      = np.logical_and(ACL1<=alpha,alpha<=(92.0*Units.deg))
    con3      = alpha>=(92.0*Units.deg)
    CL2       = np.zeros_like(alpha)
    CL2[con1] =  0
    CL2[con2] = -0.032*(alpha[con2]/Units.deg-92.0) - RCL2*((92.*Units.deg-alpha[con2])/(51.0*Units.deg))**N2
    CL2[con3] = -0.032*(alpha[con3]/Units.deg-92.0) + RCL2*((alpha[con3]-92.*Units.deg)/(51.0*Units.deg))**N2
    
    # If alpha is negative flip things for lift
    alphan    = - alpha+2*A0
    con1      = np.logical_and(0<alphan, alphan<ACL1)
    con2      = np.logical_and(ACL1<=alphan, alphan<=(92.0*Units.deg))
    con3      = alphan>=(92.0*Units.deg)
    CL2[con1] = 0.
    CL2[con2] = 0.032*(alphan[con2]/Units.deg-92.0) + RCL2*((92.*Units.deg-alphan[con2])/(51.0*Units.deg))**N2
    CL2[con3] = 0.032*(alphan[con3]/Units.deg-92.0) - RCL2*((alphan[con3]-92.*Units.deg)/(51.0*Units.deg))**N2
    
    # Equation 12a 
    con1      = np.logical_and((2*A0-ACL1)<alpha, alpha<ACL1)
    con2      = alpha>ACD1
    CD2       = 0.0 * np.ones_like(alpha)
    CD2[con1] = 0.
    CD2[con2] = CD1max[con2] + (CD2max - CD1max[con2]) * np.sin((alpha[con2]-ACD1[con2])/(np.pi/2-ACD1[con2]))
    
    # If alpha is negative flip things for drag
    alphan    = -alpha + 2*A0
    con1      = np.logical_and((2*A0-ACL1)<alphan,alphan<ACL1)
    con2      = alphan>=ACD1
    CD2[con1] = 0.
    CD2[con2] = CD1max[con2] + (CD2max - CD1max[con2]) * np.sin((alphan[con2]-ACD1[con2])/(np.pi/2-ACD1[con2]))        
        
    # Pack outputs
    wing_result = Data(
        lift_coefficient = CL2,
        drag_coefficient = CD2
        )
    
    state.conditions.aerodynamics.post_stall_coefficients[wing.tag] =  wing_result
    
    return CL2, CD2


# RCAIDE/Methods/Energy/Propulsors/Electric_Rotor_Propulsor/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .compute_electric_rotor_performance import compute_electric_rotor_performance
from .append_electric_rotor_conditions   import append_electric_rotor_conditions
from .pack_electric_rotor_residuals      import pack_electric_rotor_residuals
from .unpack_electric_rotor_unknowns     import unpack_electric_rotor_unknowns
from .append_electric_rotor_residual_and_unknown import  append_electric_rotor_residual_and_unknown


# RCAIDE/Compoments/Wings/Control_Surfaces/Aileron.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports   
from .Control_Surface import Control_Surface 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Aileron
# ----------------------------------------------------------------------------------------------------------------------  
class Aileron(Control_Surface):
    """
    A class representing an aileron control surface for aircraft roll control.

    Attributes
    ----------
    tag : str
        Unique identifier for the aileron, defaults to 'aileron'
        
    hinge_fraction : float
        Location of the hinge line as fraction of chord, defaults to 0.0
        
    sign_duplicate : float
        Sign convention for duplicate aileron deflection, defaults to -1.0
        (opposite deflection for roll control)

    Notes
    -----
    The aileron is a trailing edge control surface used primarily for roll control. 
    It inherits basic control surface functionality from the Control_Surface class 
    and adds specific attributes for aileron operation.
    
    **Definitions**

    'Hinge Fraction'
        The chordwise location of the aileron hinge line, measured from the leading 
        edge as a fraction of local chord
        
    'Sign Duplicate'
        Determines whether paired ailerons deflect in the same or opposite directions. 
        -1.0 indicates opposite deflection for roll control

    See Also
    --------
    RCAIDE.Library.Components.Wings.Control_Surfaces.Control_Surface
        Base class providing common control surface functionality
    RCAIDE.Library.Components.Wings.Control_Surfaces.Flap
        Similar trailing edge control surface used for high lift
    """ 

    def __defaults__(self):
        """
        Sets default values for the aileron attributes.
        
        Notes
        -----
        See Control_Surface.__defaults__ for additional inherited attributes.
        """
        self.tag            = 'aileron'  
        self.hinge_fraction = 0.0
        self.sign_duplicate = -1.0
        
        pass 